/**
 * Version 1.0 is to handle single variable 2DSquareCelled raster data.
 * Copyright (C) 2005 Andy Turner, CCG, University of Leeds, UK.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 */
package uk.ac.leeds.ccg.andyt.grids.core;
import gnu.trove.TDoubleHashSet;
import gnu.trove.TIntHashSet;
import java.math.BigInteger;
import java.math.BigDecimal;

/**
 * An interface to be implemented by classes that provide statistics about 
 * raster data.
 */
public interface GridStatisticsInterface {
    
    
    /**
     * For returning the number of cells with noDataValues as a BigInteger
     */
    public BigInteger getNonNoDataValueCountBigInteger();
    
    /**
     * For returning the number of cells with noDataValues as a long
     */
    public long getNonNoDataValueCountLong();
    
    
    /**
     * For returning the number of cells with noDataValues as a int
     */
    public int getNonNoDataValueCountInt();
    
    
    
    /**
     * For returning the sum of all non noDataValues as a BigDecimal
     */
    public BigDecimal getSumBigDecimal();

    /**
     * For returning the sum of all non noDataValues as a BigInteger
     */
    public BigInteger getSumBigInteger();
    
    /**
     * For returning the sum of all non noDataValues as a double
     */
    public double getSumDouble();
    
    /**
     * For returning the sum of all non noDataValues as a long
     */
    public long getSumLong();
    
    /**
     * For returning the sum of all non noDataValues as a int
     */
    public int getSumInt();
    
    
    /**
     * For returning the minimum of all non noDataValues as a BigDecimal
     */
    public BigDecimal getMinBigDecimal();
    
    /**
     * For returning the minimum of all non noDataValues as a BigInteger
     */
    public BigInteger getMinBigInteger();
    
    /**
     * For returning the minimum of all non noDataValues as a double
     */
    public double getMinDouble();
    
    /**
     * For returning the minimum of all non noDataValues as a long
     */
    public long getMinLong();
    
    /**
     * For returning the minimum of all non noDataValues as a int
     */
    public int getMinInt();
    
    /**
     * For returning the maximum of all non noDataValues as a BigDecimal
     */
    public BigDecimal getMaxBigDecimal();
    
    /**
     * For returning the maximum of all non noDataValues as a BigInteger
     */
    public BigInteger getMaxBigInteger();
    
    /**
     * For returning the minimum of all non noDataValues as a double
     */
    public double getMaxDouble();
    
    /**
     * For returning the minimum of all non noDataValues as a long
     */
    public long getMaxLong();
    
    /**
     * For returning the minimum of all non noDataValues as a int
     */
    public int getMaxInt();
    
    
    /**
     * For returning the arithmetic mean of all non noDataValues as a BigDecimal
     */
    public BigDecimal getArithmeticMeanBigDecimal( int numberOfDecimalPlaces );
    
//    /**
//     * For returning the arithmetic mean of all non noDataValues as a BigInteger
//     */
//    public BigInteger getArithmeticMeanBigInteger();
//    
//    /**
//     * For returning the arithmetic mean of all non noDataValues as a double
//     */
//    public double getArithmeticMeanDouble();
//    
//    /**
//     * For returning the arithmetic mean of all non noDataValues as a long
//     */
//    public long getArithmeticMeanLong();
//    
//    /**
//     * For returning the arithmetic mean of all non noDataValues as a int
//     */
//    public int getArithmeticMeanInt();
//      
//    /**
//     * For returning the mode of all non noDataValues as a TDoubleHashSet
//     */
//    public AbstractSet getMode();
//    
//    /**
//     * For returning the mode of all non noDataValues as a TDoubleHashSet
//     */
//    public TDoubleHashSet getModeTDoubleHashSet();
//    
//    /**
//     * For returning the mode of all non noDataValues as a TIntHashSet
//     */
//    public TIntHashSet getModeTIntHashSet();
//    
//    
//    /**
//     * For returning the standard deviation of all non noDataValues as a double
//     */
//    public double getStandardDeviationDouble();
//
//    /**
//     * For returning the geometric mean of all non noDataValues as a double
//     */
//    public double getGeometricMean();
//
//    /**
//     * For returning the arithmetic mean of all non noDataValues as a double
//     */
//    public double getHarmonicMean();
//
//    /**
//     * For returning the median of all non noDataValues as a double
//     */
//    public double getMedian();
//   
//    /**
//     * For returning the number of different values.
//     */
//    public BigInteger getDiversity();
//    
//    /**
//     * For returning all the statistics as an Object
//     */
//    public AbstractGrid2DSquareCellDoubleStatistics getStatistics();
    
}
