/**
 * Version 1.0 is to handle single variable 2DSquareCelled raster data.
 * Copyright (C) 2005 Andy Turner, CCG, University of Leeds, UK.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 */
package uk.ac.leeds.ccg.andyt.grids.process;
import java.io.File;
import java.io.BufferedReader;
import java.io.BufferedInputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.io.StreamTokenizer;
import java.io.BufferedOutputStream;
import java.io.IOException;
import java.math.BigDecimal;
import java.util.Calendar;
import java.util.HashSet;
import java.util.Iterator;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellAbstractFactory;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellDouble;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellAbstract;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellAbstract.CellID;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellAbstract.ChunkID;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellDoubleChunkAbstractFactory;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellDoubleChunkRAFFactory;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellDoubleFactory;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellDoubleChunkAbstract;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellDoubleChunk64CellMapFactory;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellDoubleChunkArrayFactory;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellDoubleChunkJAIFactory;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellDoubleChunkMapFactory;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellInt;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellIntChunk64CellMapFactory;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellIntChunkAbstract;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellIntChunkAbstractFactory;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellIntChunkArrayFactory;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellIntChunkJAIFactory;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellIntChunkMapFactory;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellIntChunkRAFFactory;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellIntFactory;
import uk.ac.leeds.ccg.andyt.grids.core.GridStatistics0;
import uk.ac.leeds.ccg.andyt.grids.core.GridStatistics1;
import uk.ac.leeds.ccg.andyt.grids.core.GridStatisticsAbstract;
import uk.ac.leeds.ccg.andyt.grids.core.OutOfMemoryErrorHandler;
import uk.ac.leeds.ccg.andyt.grids.utilities.FileCreator;
import uk.ac.leeds.ccg.andyt.grids.utilities.Utilities;

/**
 * A class holding methods for processing an individual Grid2DSquareCellAbstract or
 * multiple Grid2DSquareCellAbstracts.
 * TODO:
 * Implement a general replace method ( depricate/replace mask methods ).
 */

public class Grid2DSquareCellProcessor extends OutOfMemoryErrorHandler {
    
    /**
     * For storing the start time of the processing.
     */
    protected long startTime;
    
    /**
     * The log for recording progress and information about the processing.
     */
    protected PrintWriter log;
    
    /**
     * The log indentation (how many spaces before a log message line is output).
     */
    protected int logIndentation;
    
    /**
     * The workspace directory for the processing.
     */
    protected File workspace;
    
    /**
     * A collection of the Grid2dSquareCellAbstracts being processed
     */
    public HashSet grid2DSquareCells;
    
    /**
     * Default Grid2DSquareCellIntChunkAbstractFactory
     */
    public Grid2DSquareCellIntChunkAbstractFactory grid2DSquareCellIntChunkFactory;
    
    /**
     * Grid2DSquareCellIntChunk64CellMapFactory
     */
    public Grid2DSquareCellIntChunk64CellMapFactory grid2DSquareCellIntChunk64CellMapFactory;
    
    /**
     * Grid2DSquareCellIntChunkArrayFactory
     */
    public Grid2DSquareCellIntChunkArrayFactory grid2DSquareCellIntChunkArrayFactory;
    
    /**
     * Grid2DSquareCellIntChunkJAIFactory
     */
    public Grid2DSquareCellIntChunkJAIFactory grid2DSquareCellIntChunkJAIFactory;
    
    /**
     * Grid2DSquareCellIntChunkMapFactory
     */
    public Grid2DSquareCellIntChunkMapFactory grid2DSquareCellIntChunkMapFactory;
    
    /**
     * Grid2DSquareCellIntChunkRAFFactory
     */
    public Grid2DSquareCellIntChunkRAFFactory grid2DSquareCellIntChunkRAFFactory;
    
    /**
     * Grid2DSquareCellIntFactory
     */
    public Grid2DSquareCellIntFactory grid2DSquareCellIntFactory;
    
    /**
     * Default Grid2DSquareCellDoubleChunkAbstractFactory
     */
    public Grid2DSquareCellDoubleChunkAbstractFactory grid2DSquareCellDoubleChunkFactory;
    
    /**
     * Grid2DSquareCellDoubleChunk64CellMapFactory
     */
    public Grid2DSquareCellDoubleChunk64CellMapFactory grid2DSquareCellDoubleChunk64CellMapFactory;
    
    /**
     * Grid2DSquareCellDoubleChunkArrayFactory
     */
    public Grid2DSquareCellDoubleChunkArrayFactory grid2DSquareCellDoubleChunkArrayFactory;
    
    /**
     * Grid2DSquareCellDoubleChunkJAIFactory
     */
    public Grid2DSquareCellDoubleChunkJAIFactory grid2DSquareCellDoubleChunkJAIFactory;
    
    /**
     * Grid2DSquareCellDoubleChunkMapFactory
     */
    public Grid2DSquareCellDoubleChunkMapFactory grid2DSquareCellDoubleChunkMapFactory;
    
    /**
     * Grid2DSquareCellDoubleChunkRAFFactory
     */
    public Grid2DSquareCellDoubleChunkRAFFactory grid2DSquareCellDoubleChunkRAFFactory;
    
    /**
     * Grid2DSquareCellDoubleFactory
     */
    public Grid2DSquareCellDoubleFactory grid2DSquareCellDoubleFactory;
    
    /**
     * GridStatistics0
     */
    public GridStatistics0 gridStatistics0;
    
    /**
     * GridStatistics1
     */
    public GridStatistics1 gridStatistics1;
    
    /**
     * GridStatistics1
     */
    public GridStatisticsAbstract gridStatistics;
    
    /** Creates a new instance of Grid2DSquareCellDoubleProcessor */
    public Grid2DSquareCellProcessor() {
        //this( FileCreator.createNewFile() );
        //this( FileCreator.createNewFile( new File( System.getProperty( "tmpdir" ) ) ) );
        this( FileCreator.createNewFile( new File( System.getProperty( "java.io.tmpdir" ) ) ) );
    }
    
    /**
     * Creates a new instance of Grid2DSquareCellDoubleProcessor.
     * By default the logs are appended to the end of the log file if it exists.
     * To overwrite the log file use:
     * Grid2DSquareCellDoubleProcessor( workspace, false );
     * @param workspace
     */
    public Grid2DSquareCellProcessor( File workspace ) {
        this( workspace, true );
    }
    
    /**
     * Creates a new instance of Grid2DSquareCellDoubleProcessor. The log file
     * in workspace will be overwritten if appendToLogFile is false.
     * @param workspace
     * @param appendToLogFile
     */
    public Grid2DSquareCellProcessor( File workspace, boolean appendToLogFile ) {
        try {
            initMemoryReserve();
            this.startTime = System.currentTimeMillis();
            File logFile;
            if ( workspace.exists() ) {
                logFile = new File( workspace, "log.txt" );
                if ( ! logFile.exists() ) {
                    logFile.createNewFile();
                }
                if ( appendToLogFile ) {
                    
                }
            } else {
                workspace.mkdir();
                logFile = new File( workspace, "log.txt" );
                logFile.createNewFile();
            }
            this.workspace = workspace;
            this.log = new PrintWriter( new FileOutputStream( logFile, true ) );
            this.logIndentation = 0;
            log( this.logIndentation, "log file " + workspace.toString() + File.separator + "log.txt set up " + Calendar.getInstance().getTime().toString(), this.handleOutOfMemoryErrorFalse );
            initFactories();
            // Initialise Grid2DSquareCell container
            this.grid2DSquareCells = new HashSet();
        } catch ( IOException ioe0 ) {
            System.out.println( ioe0.getMessage() );
            ioe0.printStackTrace();
        }
    }
    
    /**
     * Initialises All Factories.
     */
    private void initFactories() {
        initChunkFactories();
        initFactories( this.workspace );
        initGridStatistics();
    }
    
    /**
     * Initialises Non Chunk Factories.
     * @param workspace The directory factories will create workspace directories in.
     */
    private void initFactories( File workspace ) {
        this.grid2DSquareCellIntFactory = new Grid2DSquareCellIntFactory( new File( workspace, "Grid2DSquareCellInt" ) );
        this.grid2DSquareCellDoubleFactory = new Grid2DSquareCellDoubleFactory( new File( workspace, "Grid2DSquareCellDouble" ) );
    }
    
    /**
     * Initialises Grid2DSquareCellChunk Factories.
     */
    private void initChunkFactories() {
        this.grid2DSquareCellIntChunk64CellMapFactory = new Grid2DSquareCellIntChunk64CellMapFactory();
        this.grid2DSquareCellIntChunkArrayFactory = new Grid2DSquareCellIntChunkArrayFactory();
        this.grid2DSquareCellIntChunkJAIFactory = new Grid2DSquareCellIntChunkJAIFactory();
        this.grid2DSquareCellIntChunkMapFactory = new Grid2DSquareCellIntChunkMapFactory();
        this.grid2DSquareCellIntChunkRAFFactory = new Grid2DSquareCellIntChunkRAFFactory();
        this.grid2DSquareCellIntChunkFactory = grid2DSquareCellIntChunkArrayFactory;
        this.grid2DSquareCellDoubleChunk64CellMapFactory = new Grid2DSquareCellDoubleChunk64CellMapFactory();
        this.grid2DSquareCellDoubleChunkArrayFactory = new Grid2DSquareCellDoubleChunkArrayFactory();
        this.grid2DSquareCellDoubleChunkJAIFactory = new Grid2DSquareCellDoubleChunkJAIFactory();
        this.grid2DSquareCellDoubleChunkMapFactory = new Grid2DSquareCellDoubleChunkMapFactory();
        this.grid2DSquareCellDoubleChunkRAFFactory = new Grid2DSquareCellDoubleChunkRAFFactory();
        this.grid2DSquareCellDoubleChunkFactory = grid2DSquareCellDoubleChunkArrayFactory;
    }
    
    /**
     * Initialises GridStatistics.
     */
    private void initGridStatistics() {
        this.gridStatistics0 = new GridStatistics0();
        this.gridStatistics1 = new GridStatistics1();
        this.gridStatistics = gridStatistics1;
    }
    
    /**
     * Returns a copy of this.startTime.
     */
    public long getTime0() {
        return this.startTime;
    }
    
    /**
     * Returns a copy of this.workspace
     */
    public File getWorkspace( boolean handleOutOfMemoryError ) {
        try {
            return new File( this.workspace.toString() );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellChunk( handleOutOfMemoryError );
                File result = getWorkspace( handleOutOfMemoryError );
                initMemoryReserve();
            }
            throw oome0;
        }
    }
    
    /**
     * Returns this.grid2DSquareCells
     */
    protected HashSet getGrid2DSquareCells() {
        return this.grid2DSquareCells;
    }
    
    /**
     * Attempts to swap to file and clear from the cache a chunk in
     * this.grid2DSquareCells
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final void swapToFileGrid2DSquareCellChunk(
            boolean handleOutOfMemoryError ) {
        try {
            swapToFileGrid2DSquareCellChunk();
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellChunk();
                initMemoryReserve();
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Attempts to swap to file and clear from the cache a Grid2DSquareCellChunk
     * and all coincident Grid2DSquareCellChunks in this.grid2DSquareCells.
     * WARNING!!! No test for intersection is done. It is assumed that all
     * Grid2DSquareCellAbstract in this.grid2DSquareCells have the same structure
     * and dimensions.
     */
    private final void swapToFileGrid2DSquareCellChunk() {
        Iterator iterator = this.grid2DSquareCells.iterator();
        Grid2DSquareCellAbstract grid2DSquareCell = 
                ( Grid2DSquareCellAbstract ) iterator.next();
        grid2DSquareCell.swapToFileGrid2DSquareCellsGrid2DSquareCellChunk( true );
    }
    
    
    /**
     * Attempts to swap to file and clear from cache seriailsed version of all
     * this.grid2DSquareCells chunks.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final void swapToFileGrid2DSquareCellChunks(
            boolean handleOutOfMemoryError ) {
        try {
            swapToFileGrid2DSquareCellChunks();
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellChunk();
                initMemoryReserve();
                swapToFileGrid2DSquareCellChunks( handleOutOfMemoryError );
            }
            throw oome0;
        }
    }
    
    public final void swapToFileGrid2DSquareCellChunks() {
        Iterator iterator = this.grid2DSquareCells.iterator();
        Grid2DSquareCellAbstract grid2DSquareCell = 
                ( Grid2DSquareCellAbstract ) iterator.next();
        grid2DSquareCell.swapToFileGrid2DSquareCellsGrid2DSquareCellChunks( true );
    }
    
//    /**
//     * Attempts to swap to file and clear from cache seriailsed version of all
//     * this.grid2DSquareCell chunks from (cri0, cci0) to (cri1, cci1). Generally
//     * this is most appropriate only if all grids in this.grid2DSquareCells are
//     * coincident.
//     * @param cri0
//     * @param cci0
//     * @param cri1
//     * @param cci1
//     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
//     *   in this method then swap operations are initiated prior to retrying.
//     *   If false then OutOfMemoryErrors are caught and thrown.
//     * @param reinitialiseMemoryReserve If true then this.memoryReserve is initialised
//     *   prior to returning from this method.
//     */
//    public final void swapToFileGrid2DSquareCellChunks( 
//            int cri0, 
//            int cci0, 
//            int cri1, 
//            int cci1, 
//            boolean handleOutOfMemoryError ) {
//        try {
//            if ( true ) {
//                Iterator iterator = this.grid2DSquareCells.iterator();
//                Grid2DSquareCellAbstract grid2DSquareCell;
//                while ( iterator.hasNext() ) {
//                    grid2DSquareCell = ( Grid2DSquareCellAbstract ) iterator.next();
//                    grid2DSquareCell.swapToFileGrid2DSquareCellChunks( cri0, cci0, cri1, cci1, false , handleOutOfMemoryError);
//                }
//            }
//            if ( reInitialiseMemoryReserve ) {
//                initMemoryReserve();
//            }
//        } catch ( java.lang.OutOfMemoryError oome0 ) {
//            if ( handleOutOfMemoryError ) {
//                clearMemoryReserve();
//                swapToFileGrid2DSquareCellChunks( cri0, cci0, cri1, cci1, handleOutOfMemoryError, reInitialiseMemoryReserve );
//                if ( reInitialiseMemoryReserve ) {
//                    try {
//                        initMemoryReserve();
//                    } catch ( java.lang.OutOfMemoryError oome1 ) {
//                        System.out.println( oome1.getMessage() );
//                        oome1.printStackTrace();
//                        throw oome1;
//                        //throw new OutOfMemoryError( this.getClass().toString() + ".swapToFileChunks( cri0( " + cri0 + " ), cci0( " + cci0 + " ), cri1( " + cri1 + " ), cci1( " + cci1 + " ), handleOutOfMemoryError( " + handleOutOfMemoryError + "), reInitialiseMemoryReserve( " + reInitialiseMemoryReserve + " ) ), problem re-initialising memory!" );
//                    }
//                }
//                return;
//            }
//            throw oome0;
//        }
//    }
    
    /**
     * Changes workspace to that passed in if it can be created. By default this
     * does not copy the logfile from the existing workspace when it sets up a
     * log in the new location. To do this use:
     * setWorkspace( workspace, true )
     * @param workspace The workspace to change to.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public void setWorkspace( 
            File workspace, 
            boolean handleOutOfMemoryError ) {
        try {
            this.workspace = workspace;
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellChunk( true );
                setWorkspace( workspace, handleOutOfMemoryError );
                initMemoryReserve();
                return;
            }
            throw oome0;
        }
    }
    
    /**
     * Changes workspace to that passed in if it can be created. If copyLogFile
     * is true, this copies the logfile from the existing workspace and sets up
     * the log to append in the new location.
     * @param workspace The workspace to change to.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public void setWorkspace( File workspace, boolean copyLogFile, boolean handleOutOfMemoryError ) {
        try{
            boolean mkdirSuccess = false;
            File newLog = new File( workspace, "log.txt" );
            try {
                if ( ! workspace.exists() ) {
                    mkdirSuccess = workspace.mkdir();
                    if ( copyLogFile ) {
                        copyAndSetUpNewLog( newLog, this.handleOutOfMemoryErrorFalse );
                    } else {
                        newLog.createNewFile();
                        this.log = new PrintWriter( new FileOutputStream( newLog, true ) );
                    }
                } else {
                    if ( ! newLog.exists() ) {
                        newLog.createNewFile();
                        this.log = new PrintWriter( new FileOutputStream( newLog, true ) );
                    } else {
                        if ( copyLogFile ) {
                            this.log = new PrintWriter( new FileOutputStream( newLog, true ) );
                        } else {
                            this.log = new PrintWriter( new FileOutputStream( newLog, false ) );
                        }
                    }
                }
            } catch ( IOException ioe0 ) {
                ioe0.printStackTrace();
            }
            this.workspace = workspace;
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellChunk( true );
                setWorkspace( workspace, copyLogFile, handleOutOfMemoryError );
                initMemoryReserve();
                return;
            }
            throw oome0;
        }
    }
    
    /**
     * Copies and sets up a new log.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public void copyAndSetUpNewLog( File newLog, boolean handleOutOfMemoryError ) throws IOException {
        try {
            this.log.flush();
            this.log.close();
            File workspace = getWorkspace(this.handleOutOfMemoryErrorTrue);
            File oldLog = new File( workspace, "log.txt" );
            BufferedInputStream bis = new BufferedInputStream( new FileInputStream( oldLog ) );
            BufferedOutputStream bos = new BufferedOutputStream( new FileOutputStream( newLog ) );
            for ( int i = 0; i < oldLog.length(); i ++ ) {
                bos.write( bis.read() );
            }
            bos.flush();
            bos.close();
            bis.close();
            this.log = new PrintWriter( new FileOutputStream( newLog, true ) );
            log( "log file copied from " + oldLog.toString() + " " + Calendar.getInstance().toString(), this.handleOutOfMemoryErrorFalse );
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellChunk( true );
                copyAndSetUpNewLog( newLog, handleOutOfMemoryError );
                initMemoryReserve();
                return;
            }
            throw oome0;
        }
    }
    
    /**
     * Writes string to log file and the console (standard output)
     * @param string The message to log.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public void log( String string, boolean handleOutOfMemoryError ) {
        log( this.logIndentation, string, handleOutOfMemoryError );
    }
    
    /**
     * Writes string to log file and the console (standard output) indenting
     * string by logIndentation amount of white-space.
     * @param logIndentation The indentation of string.
     * @param string The message to log.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public void log( 
            int logIndentation, 
            String string, 
            boolean handleOutOfMemoryError ) {
        try {
            if ( string.endsWith( "}" ) ) {
                logIndentation --;
                this.logIndentation --;
            }
            for ( int i = 0; i < logIndentation; i ++ ) {
                System.out.print( " " );
                this.log.write( " " );
            }
            System.out.println( string );
            this.log.write( string );
            this.log.write( System.getProperty( "line.separator" ) );
            this.log.flush();
            if ( string.endsWith( "{" ) ) {
                this.logIndentation ++;
            }
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellChunk( true );
                log( logIndentation, string, handleOutOfMemoryError );
                initMemoryReserve();
                return;
            }
            throw oome0;
        }
    }
    
    /**
     * Returns the distance between a pair of coordinates
     * @param x1 - the x coordinte of one point
     * @param y1 - the y coordinte of one point
     * @param x2 - the x coordinte of another point
     * @param y2 - the y coordinte of another point
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final double distance( 
            double x1, 
            double y1, 
            double x2, 
            double y2, 
            boolean handleOutOfMemoryError ) {
        try {
            return Utilities.distance( x1, y1, x2, y2 );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellChunk( true );
                double distance = distance( 
                        x1, y1, x2, y2, 
                        handleOutOfMemoryError );
                initMemoryReserve();
                return distance;
            }
            throw oome0;
        }
    }
    
    /**
     * Returns the clockwise angle in radians to the y axis of the line from 
     * (x1,y1) to (x2,y2).
     * @param x1 - the x coordinte of one point
     * @param y1 - the y coordinte of one point
     * @param x2 - the x coordinte of another point
     * @param y2 - the y coordinte of another point
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final double angle( 
            double x1, 
            double y1, 
            double x2, 
            double y2, 
            boolean handleOutOfMemoryError ) {
        try {
            return Utilities.angle( 
                    x1, 
                    y1, 
                    x2, 
                    y2 );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellChunk( handleOutOfMemoryError );
                double distance = angle( 
                        x1,
                        y1, 
                        x2,
                        y2, 
                        handleOutOfMemoryError );
                initMemoryReserve();
                return distance;
            }
            throw oome0;
        }
    }
    
    /**
     * Returns a new Grid2DSquareCellAbstract which is of the same type and chunk stucture as
     * grid2DSquareCellMask.
     * @param grid2DSquareCellToMask The Grid2DSquareCellAbstract that is copied for masking.
     * @param grid2DSquareCellMask The Grid2DSquareCellAbstract to use as a mask.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * Warning!!! The grid2DSquareCellToMask and grid2DSquareCellMask are assumed to be
     * coincident and have the same origin.
     */
    public Grid2DSquareCellAbstract mask(
            Grid2DSquareCellAbstract grid2DSquareCellToMask,
            Grid2DSquareCellAbstract grid2DSquareCellMask,
            GridStatisticsAbstract resultGridStatistics,
            File resultDirectory,
            boolean handleOutOfMemoryError ) {
        try {
            this.grid2DSquareCells.add( grid2DSquareCellToMask );
            this.grid2DSquareCells.add( grid2DSquareCellMask );
            int chunkNrows = grid2DSquareCellToMask.getChunkNrows(
                    this.handleOutOfMemoryErrorFalse );
            int chunkNcols = grid2DSquareCellToMask.getChunkNcols(
                    this.handleOutOfMemoryErrorFalse );
            long nrows = grid2DSquareCellToMask.getNrows( handleOutOfMemoryError );
            long ncols = grid2DSquareCellToMask.getNcols( handleOutOfMemoryError );
            ChunkID chunkID;
            int thisChunkNrows;
            int thisChunkNcols;
            int chunkRowIndex;
            int chunkColIndex;
            int chunkCellRowIndex;
            int chunkCellColIndex;
            if ( grid2DSquareCellToMask.getClass() == Grid2DSquareCellInt.class ) {
                Grid2DSquareCellInt grid2DSquareCellToMaskInt = 
                        ( Grid2DSquareCellInt ) grid2DSquareCellToMask;
                Grid2DSquareCellIntFactory grid2DSquareCellIntFactory =
                        new Grid2DSquareCellIntFactory( 
                        resultDirectory, 
                        chunkNrows, 
                        chunkNcols, 
                        this.grid2DSquareCellIntChunkArrayFactory );
                Grid2DSquareCellInt result = 
                        ( Grid2DSquareCellInt ) grid2DSquareCellIntFactory.create( 
                        resultGridStatistics, 
                        resultDirectory, 
                        grid2DSquareCellToMask, 
                        this.grid2DSquareCellIntChunkArrayFactory, 
                        0L, 
                        0L, 
                        nrows -1L, 
                        ncols - 1L, 
                        grid2DSquareCells, 
                        handleOutOfMemoryError );
                int resultNoDataValue = result.getNoDataValue( handleOutOfMemoryError );
                if ( grid2DSquareCellMask.getClass() == Grid2DSquareCellInt.class ) {
                    Grid2DSquareCellInt grid2DSquareCellMaskInt = 
                            ( Grid2DSquareCellInt ) grid2DSquareCellMask;
                    int maskNoDataValue = grid2DSquareCellMaskInt.getNoDataValue( 
                            handleOutOfMemoryError );
                    int value;
                    Iterator iterator = grid2DSquareCellMaskInt.iterator( 
                            handleOutOfMemoryError );
                    Grid2DSquareCellIntChunkAbstract grid2DSquareCellMaskIntChunk;
                    Grid2DSquareCellIntChunkAbstract resultChunk;
                    while ( iterator.hasNext() ) {
                        grid2DSquareCellMaskIntChunk = 
                                ( Grid2DSquareCellIntChunkAbstract ) iterator.next();
                        chunkID = grid2DSquareCellMaskIntChunk.getChunkID();
                        thisChunkNrows = grid2DSquareCellMaskInt.getChunkNrows( 
                                
                                chunkID, handleOutOfMemoryError );
                        thisChunkNcols = grid2DSquareCellMaskInt.getChunkNcols( 
                                chunkID, 
                                handleOutOfMemoryError );
                        chunkRowIndex = chunkID.getChunkRowIndex();
                        chunkColIndex = chunkID.getChunkColIndex();
                        for ( chunkCellRowIndex = 0; chunkCellRowIndex < chunkNrows; chunkCellRowIndex ++ ) {
                            for ( chunkCellColIndex = 0; chunkCellColIndex < chunkNcols; chunkCellColIndex ++ ) {
                                value = grid2DSquareCellMaskIntChunk.getCell( 
                                        chunkCellRowIndex, 
                                        chunkCellColIndex, 
                                        maskNoDataValue );
                                if ( value == maskNoDataValue ) {
                                    result.setCell( 
                                            ( ( long ) chunkRowIndex * ( long ) chunkNrows ) + ( long ) chunkCellRowIndex,
                                            ( ( long ) chunkColIndex * ( long ) chunkNcols ) + ( long ) chunkCellColIndex, 
                                            resultNoDataValue, 
                                            handleOutOfMemoryError );
                                }
                            }
                        }
                    }
                } else {
                    // ( grid2DSquareCellMask.getClass() == Grid2DSquareCellDouble.class )
                    Grid2DSquareCellDouble grid2DSquareCellMaskDouble = 
                            ( Grid2DSquareCellDouble ) grid2DSquareCellMask;
                    double maskNoDataValue = grid2DSquareCellMaskDouble.getNoDataValue(
                            handleOutOfMemoryError );
                    double value;
                    Iterator iterator = grid2DSquareCellMaskDouble.iterator( 
                            handleOutOfMemoryError );
                    Grid2DSquareCellDoubleChunkAbstract grid2DSquareCellMaskDoubleChunk;
                    Grid2DSquareCellIntChunkAbstract resultChunk;
                    while ( iterator.hasNext() ) {
                        grid2DSquareCellMaskDoubleChunk = 
                                ( Grid2DSquareCellDoubleChunkAbstract ) iterator.next();
                        chunkID = grid2DSquareCellMaskDoubleChunk.getChunkID();
                        thisChunkNrows = grid2DSquareCellMaskDouble.getChunkNrows( 
                                chunkID, 
                                handleOutOfMemoryError );
                        thisChunkNcols = grid2DSquareCellMaskDouble.getChunkNcols( 
                                chunkID, 
                                handleOutOfMemoryError );
                        chunkRowIndex = chunkID.getChunkRowIndex();
                        chunkColIndex = chunkID.getChunkColIndex();
                        for ( chunkCellRowIndex = 0; chunkCellRowIndex < chunkNrows; chunkCellRowIndex ++ ) {
                            for ( chunkCellColIndex = 0; chunkCellColIndex < chunkNcols; chunkCellColIndex ++ ) {
                                value = grid2DSquareCellMaskDoubleChunk.getCell( 
                                        chunkCellRowIndex, 
                                        chunkCellColIndex, 
                                        maskNoDataValue );
                                if ( value == maskNoDataValue ) {
                                    result.setCell( 
                                            ( ( long ) chunkRowIndex * ( long ) chunkNrows ) + ( long ) chunkCellRowIndex, 
                                            ( ( long ) chunkColIndex * ( long ) chunkNcols ) + ( long ) chunkCellColIndex, 
                                            resultNoDataValue,
                                            handleOutOfMemoryError );
                                }
                            }
                        }
                    }
                }
                return result;
            } else {
                //( grid2DSquareCellToMask.getClass() == Grid2DSquareCellDouble.class ) {
                double resultNoDataValue = 
                        ( ( Grid2DSquareCellDouble ) grid2DSquareCellToMask ).getNoDataValue(
                        handleOutOfMemoryError );
                Grid2DSquareCellInt grid2DSquareCellToMaskInt = 
                        ( Grid2DSquareCellInt ) grid2DSquareCellToMask;
                Grid2DSquareCellDoubleFactory grid2DSquareCellDoubleFactory = 
                        new Grid2DSquareCellDoubleFactory( 
                        resultDirectory, 
                        chunkNrows, 
                        chunkNcols, 
                        this.grid2DSquareCellDoubleChunkArrayFactory, 
                        resultNoDataValue );
                Grid2DSquareCellDouble result = 
                        ( Grid2DSquareCellDouble ) grid2DSquareCellDoubleFactory.create( 
                        resultGridStatistics, 
                        resultDirectory, 
                        grid2DSquareCellToMask, 
                        this.grid2DSquareCellDoubleChunkFactory, 
                        0L, 
                        0L,
                        nrows -1L,
                        ncols - 1L, 
                        grid2DSquareCells, 
                        handleOutOfMemoryError );
                if ( grid2DSquareCellMask.getClass() == Grid2DSquareCellInt.class ) {
                    Grid2DSquareCellInt grid2DSquareCellMaskInt = 
                            ( Grid2DSquareCellInt ) grid2DSquareCellMask;
                    int maskNoDataValue = grid2DSquareCellMaskInt.getNoDataValue( 
                            handleOutOfMemoryError );
                    int value;
                    Iterator iterator = grid2DSquareCellMaskInt.iterator( 
                            handleOutOfMemoryError );
                    Grid2DSquareCellIntChunkAbstract grid2DSquareCellMaskIntChunk;
                    Grid2DSquareCellIntChunkAbstract resultChunk;
                    while ( iterator.hasNext() ) {
                        grid2DSquareCellMaskIntChunk = 
                                ( Grid2DSquareCellIntChunkAbstract ) iterator.next();
                        chunkID = grid2DSquareCellMaskIntChunk.getChunkID();
                        thisChunkNrows = grid2DSquareCellMaskInt.getChunkNrows( 
                                chunkID, 
                                handleOutOfMemoryError );
                        thisChunkNcols = grid2DSquareCellMaskInt.getChunkNcols( 
                                chunkID, 
                                handleOutOfMemoryError );
                        chunkRowIndex = chunkID.getChunkRowIndex();
                        chunkColIndex = chunkID.getChunkColIndex();
                        for ( chunkCellRowIndex = 0; chunkCellRowIndex < chunkNrows; chunkCellRowIndex ++ ) {
                            for ( chunkCellColIndex = 0; chunkCellColIndex < chunkNcols; chunkCellColIndex ++ ) {
                                value = grid2DSquareCellMaskIntChunk.getCell( 
                                        chunkCellRowIndex, 
                                        chunkCellColIndex, 
                                        maskNoDataValue );
                                if ( value == maskNoDataValue ) {
                                    result.setCell( 
                                            ( ( long ) chunkRowIndex * ( long ) chunkNrows ) + ( long ) chunkCellRowIndex, 
                                            ( ( long ) chunkColIndex * ( long ) chunkNcols ) + ( long ) chunkCellColIndex, 
                                            resultNoDataValue, handleOutOfMemoryError );
                                }
                            }
                        }
                    }
                } else {
                    // ( grid2DSquareCellMask.getClass() == Grid2DSquareCellDouble.class )
                    Grid2DSquareCellDouble grid2DSquareCellMaskDouble = 
                            ( Grid2DSquareCellDouble ) grid2DSquareCellMask;
                    double maskNoDataValue = grid2DSquareCellMaskDouble.getNoDataValue(
                            handleOutOfMemoryError );
                    double value;
                    Iterator iterator = grid2DSquareCellMaskDouble.iterator( 
                            handleOutOfMemoryError );
                    Grid2DSquareCellDoubleChunkAbstract grid2DSquareCellMaskDoubleChunk;
                    Grid2DSquareCellIntChunkAbstract resultChunk;
                    while ( iterator.hasNext() ) {
                        grid2DSquareCellMaskDoubleChunk = 
                                ( Grid2DSquareCellDoubleChunkAbstract ) iterator.next();
                        chunkID = grid2DSquareCellMaskDoubleChunk.getChunkID();
                        thisChunkNrows = grid2DSquareCellMaskDouble.getChunkNrows( 
                                chunkID, 
                                handleOutOfMemoryError );
                        thisChunkNcols = grid2DSquareCellMaskDouble.getChunkNcols( 
                                chunkID, 
                                handleOutOfMemoryError );
                        chunkRowIndex = chunkID.getChunkRowIndex();
                        chunkColIndex = chunkID.getChunkColIndex();
                        for ( chunkCellRowIndex = 0; chunkCellRowIndex < chunkNrows; chunkCellRowIndex ++ ) {
                            for ( chunkCellColIndex = 0; chunkCellColIndex < chunkNcols; chunkCellColIndex ++ ) {
                                value = grid2DSquareCellMaskDoubleChunk.getCell( 
                                        chunkCellRowIndex, 
                                        chunkCellColIndex, 
                                        maskNoDataValue );
                                if ( value == maskNoDataValue ) {
                                    result.setCell( 
                                            ( ( long ) chunkRowIndex * ( long ) chunkNrows ) + ( long ) chunkCellRowIndex, 
                                            ( ( long ) chunkColIndex * ( long ) chunkNcols ) + ( long ) chunkCellColIndex, 
                                            resultNoDataValue, 
                                            handleOutOfMemoryError );
                                }
                            }
                        }
                    }
                }
                return result;
            }
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellChunks( handleOutOfMemoryError );
                Grid2DSquareCellAbstract result = mask( 
                        grid2DSquareCellToMask, 
                        grid2DSquareCellMask, 
                        resultGridStatistics, 
                        resultDirectory, 
                        handleOutOfMemoryError );
                initMemoryReserve();
                return result;
            }
            throw oome0;
        }
    }
    
    /**
     * If grid2DSquareCellAbstractFactory is null then grid2DSquareCell is directly masked and
     * returned, otherwise a new Grid2DSquareCellAbstract result is contructed using
     * grid2DSquareCellAbstractFactory. Essentially this is a copy of grid2DSquareCell, masked
     * using mask. Masking involves setting result cell values to it's noDataValue if the cells
     * are conicident with mask cells with mask.noDataValue value.
     * Warning: If returned Grid2DSquareCellAbstract is not same class as grid2DSquareCell some
     * generalisation may have occured in addition to masking.
     * @param grid2DSquareCell The Grid2DSquareCellAbstract to be masked.
     * @param mask The Grid2DSquareCellAbstract mask.
     * @param grid2DSquareCellAbstractFactory the Grid2DSquareCellAbstractFactory
     *   used to generate result.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellAbstract mask(
            Grid2DSquareCellAbstract grid2DSquareCell,
            Grid2DSquareCellAbstract mask,
            Grid2DSquareCellAbstractFactory grid2DSquareCellAbstractFactory,
            boolean handleOutOfMemoryError ) {
        try {
            if ( grid2DSquareCellAbstractFactory == null ) {
                // If grid2DSquareCell and mask have coincident Grid2DSquareCellAbstractChunks
                // processesing by iterating over these offers greater likelihood  of not
                // running OutOfMemory.
                int grid2DSquareCellChunkNrows = grid2DSquareCell.getChunkNrows( handleOutOfMemoryError );
                int grid2DSquareCellChunkNcols = grid2DSquareCell.getChunkNcols( handleOutOfMemoryError );
                int maskChunkNrows = mask.getChunkNrows( handleOutOfMemoryError );
                int maskChunkNcols = mask.getChunkNcols( handleOutOfMemoryError );
                // Check for coincidence of grid2DSquareCell and mask Grid2DSquareCellAbstractChunks
            } else {
                if ( grid2DSquareCell.getClass() == Grid2DSquareCellInt.class ) {
                    Grid2DSquareCellIntFactory grid2DSquareCellIntFactory = ( Grid2DSquareCellIntFactory ) grid2DSquareCellAbstractFactory;
                    Grid2DSquareCellInt result = ( Grid2DSquareCellInt ) grid2DSquareCellIntFactory.create( grid2DSquareCell );
                    
                } else {
                    if ( grid2DSquareCell.getClass() == Grid2DSquareCellDouble.class ) {
                        Grid2DSquareCellDoubleFactory grid2DSquareCellDoubleFactory = ( Grid2DSquareCellDoubleFactory ) grid2DSquareCellAbstractFactory;
                        return grid2DSquareCellDoubleFactory.create( grid2DSquareCell );
                    }
                }
            }
            //result.setName( grid2DSquareCell.getName() + "_mask", handleOutOfMemoryError );
            return null;
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                Grid2DSquareCellAbstract result = mask( grid2DSquareCell, mask, grid2DSquareCellAbstractFactory, handleOutOfMemoryError );
                try {
                    initMemoryReserve();
                } catch ( java.lang.OutOfMemoryError oome1 ) {
                    return result;
                }
            }
            throw oome0;
        }
    }
    
    /**
     * Returns grid2DSquareCell with the values of cells in the range [min,max] set to its
     * noDataValue. (Existing noDataValue cells in grid remain as noDataValue.)
     * @param grid2DSquareCell the Grid2DSquareCellDouble to be masked
     * @param min the minimum value in the range
     * @param max the maximum value in the range
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellAbstract mask(
            Grid2DSquareCellAbstract grid2DSquareCell,
            double min, double max,
            boolean handleOutOfMemoryError ) {
        try {
            this.grid2DSquareCells.add( grid2DSquareCell );
            long ncols = grid2DSquareCell.getNcols(handleOutOfMemoryError);
            long nrows = grid2DSquareCell.getNrows(handleOutOfMemoryError);
            int chunkCellRowIndex;
            int chunkCellColIndex;
            int chunkNrows;
            int chunkNcols;
            int chunkRowIndex;
            int chunkColIndex;
            //int nChunkRows = grid2DSquareCell.getNChunkRows( handleOutOfMemoryError );
            //int nChunkCols = grid2DSquareCell.getNChunkCols( handleOutOfMemoryError );
            ChunkID chunkID;
            if ( grid2DSquareCell.getClass() == Grid2DSquareCellInt.class ) {
                Grid2DSquareCellInt grid2DSquareCellInt = ( Grid2DSquareCellInt ) grid2DSquareCell;
                int noDataValue = grid2DSquareCellInt.getNoDataValue( true );
                int value;
                Iterator iterator = grid2DSquareCellInt.iterator( handleOutOfMemoryError );
                Grid2DSquareCellIntChunkAbstract grid2DSquareCellIntChunk;
                //Iterator grid2DSquareCellChunkIterator;
                while ( iterator.hasNext() ) {
                    grid2DSquareCellIntChunk = ( Grid2DSquareCellIntChunkAbstract ) iterator.next();
                    chunkID = grid2DSquareCellIntChunk.getChunkID();
                    chunkNrows = grid2DSquareCellInt.getChunkNrows( chunkID, handleOutOfMemoryError );
                    chunkNcols = grid2DSquareCellInt.getChunkNcols( chunkID, handleOutOfMemoryError );
                    chunkRowIndex = chunkID.getChunkRowIndex();
                    chunkColIndex = chunkID.getChunkColIndex();
                    for ( chunkCellRowIndex = 0; chunkCellRowIndex < chunkNrows; chunkCellRowIndex ++ ) {
                        for ( chunkCellColIndex = 0; chunkCellColIndex < chunkNcols; chunkCellColIndex ++ ) {
                            value = grid2DSquareCellInt.getCell( grid2DSquareCellIntChunk, chunkRowIndex, chunkColIndex, chunkCellRowIndex, chunkCellColIndex, handleOutOfMemoryError );
                            if ( value >= min && value <= max ) {
                                grid2DSquareCellInt.setCell( grid2DSquareCellIntChunk, chunkRowIndex, chunkColIndex, chunkCellRowIndex, chunkCellColIndex, noDataValue, handleOutOfMemoryError );
                            }
                        }
                    }
                }
            } else {
                // ( grid2DSquareCell.getClass() == Grid2DSquareCellDouble.class )
                Grid2DSquareCellDouble grid2DSquareCellDouble = ( Grid2DSquareCellDouble ) grid2DSquareCell;
                double noDataValue = grid2DSquareCellDouble.getNoDataValue(handleOutOfMemoryError);
                double value;
                Iterator iterator = grid2DSquareCellDouble.iterator( handleOutOfMemoryError );
                Grid2DSquareCellDoubleChunkAbstract grid2DSquareCellDoubleChunk;
                //Iterator grid2DSquareCellChunkIterator;
                while ( iterator.hasNext() ) {
                    grid2DSquareCellDoubleChunk = ( Grid2DSquareCellDoubleChunkAbstract ) iterator.next();
                    chunkID = grid2DSquareCellDoubleChunk.getChunkID();
                    chunkNrows = grid2DSquareCellDouble.getChunkNrows( chunkID, handleOutOfMemoryError );
                    chunkNcols = grid2DSquareCellDouble.getChunkNcols( chunkID, handleOutOfMemoryError );
                    chunkRowIndex = chunkID.getChunkRowIndex();
                    chunkColIndex = chunkID.getChunkColIndex();
                    for ( chunkCellRowIndex = 0; chunkCellRowIndex < chunkNrows; chunkCellRowIndex ++ ) {
                        for ( chunkCellColIndex = 0; chunkCellColIndex < chunkNcols; chunkCellColIndex ++ ) {
                            value = grid2DSquareCellDouble.getCell( grid2DSquareCellDoubleChunk, chunkRowIndex, chunkColIndex, chunkCellRowIndex, chunkCellColIndex, handleOutOfMemoryError );
                            if ( value >= min && value <= max ) {
                                grid2DSquareCellDouble.setCell( grid2DSquareCellDoubleChunk, chunkRowIndex, chunkColIndex, chunkCellRowIndex, chunkCellColIndex, noDataValue, handleOutOfMemoryError );
                            }
                        }
                    }
                }
            }
            //grid.setName( grid.getName() + "_mask" );
            return grid2DSquareCell;
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                mask( grid2DSquareCell, min, max, handleOutOfMemoryError );
                try {
                    initMemoryReserve();
                } catch ( java.lang.OutOfMemoryError oome1 ) {
                    return grid2DSquareCell;
                }
            }
            throw oome0;
        }
    }
    
    /**
     * Returns grid with the value of cells that's centroids intersect the
     * rectangle given by (xmin,ymin,xmax,ymax) set to its noDataValue.
     * @param grid2DSquareCell the Grid2DSquareCellAbstract to be masked
     * @param xmin the minimum x-coordinate of the masking rectangle
     * @param ymin the minimum y-coordinate of the masking rectangle
     * @param xmax the maximum x-coordinate of the masking rectangle
     * @param ymax the maximum y-coordinate of the masking rectangle
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellAbstract mask(
            Grid2DSquareCellAbstract grid2DSquareCell,
            double xmin, double ymin, double xmax, double ymax,
            boolean handleOutOfMemoryError ) {
        //        try {
        return mask( grid2DSquareCell, grid2DSquareCell.getCellRowIndex( ymax, handleOutOfMemoryError ), grid2DSquareCell.getCellColIndex( xmin, handleOutOfMemoryError ), grid2DSquareCell.getCellRowIndex( ymin, handleOutOfMemoryError ), grid2DSquareCell.getCellColIndex( xmax, handleOutOfMemoryError ), handleOutOfMemoryError );
        //        } catch ( java.lang.OutOfMemoryError oome0 ) {
        //            if ( handleOutOfMemoryError ) {
        //                clearMemoryReserve();
        //                mask( grid2DSquareCell, xmin, ymin, xmax, ymax, handleOutOfMemoryError );
        //                try {
        //                    initMemoryReserve();
        //                } catch ( java.lang.OutOfMemoryError oome1 ) {
        //                    return grid2DSquareCell;
        //                }
        //            }
        //            throw oome0;
        //        }
    }
    
    /**
     * Returns grid with the value of cells that intersect rectangle given by
     * (startRowIndex,startColIndex,endRowIndex,endColIndex) set to its
     * noDataValue.
     * @param grid2DSquareCell the Grid2DSquareCellDouble to be masked
     * @param startRowIndex the index of the first row to be masked
     * @param startColIndex the index of the first column to be masked
     * @param endRowIndex the index of the final row to be masked
     * @param endColIndex the index of the final column to be masked
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellAbstract mask(
            Grid2DSquareCellAbstract grid2DSquareCell,
            long startRowIndex, long startColIndex,
            long endRowIndex, long endColIndex,
            boolean handleOutOfMemoryError ) {
        try {
            Grid2DSquareCellAbstract result;
            long rowIndex;
            long colIndex;
            if ( grid2DSquareCell.getClass() == Grid2DSquareCellInt.class ) {
                Grid2DSquareCellInt grid2DSquareCellInt = ( Grid2DSquareCellInt ) grid2DSquareCell;
                int noDataValue = grid2DSquareCellInt.getNoDataValue( true );
                int value;
                for ( rowIndex = startRowIndex; rowIndex <= endRowIndex; rowIndex ++ ) {
                    for ( colIndex = startColIndex; colIndex <= endColIndex; colIndex ++ ) {
                        grid2DSquareCellInt.setCell( rowIndex, colIndex, noDataValue, this.handleOutOfMemoryErrorTrue );
                    }
                }
                result = grid2DSquareCellInt;
            } else {
                // ( grid2DSquareCell.getClass() == Grid2DSquareCellDouble.class )
                Grid2DSquareCellDouble grid2DSquareCellDouble = ( Grid2DSquareCellDouble ) grid2DSquareCell;
                double noDataValue = grid2DSquareCellDouble.getNoDataValue(handleOutOfMemoryError);
                double value;
                for ( rowIndex = startRowIndex; rowIndex <= endRowIndex; rowIndex ++ ) {
                    for ( colIndex = startColIndex; colIndex <= endColIndex; colIndex ++ ) {
                        grid2DSquareCellDouble.setCell( rowIndex, colIndex, noDataValue, this.handleOutOfMemoryErrorTrue );
                    }
                }
                result = grid2DSquareCellDouble;
            }
            return result;
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                mask( grid2DSquareCell, startRowIndex, startColIndex, endRowIndex, endColIndex, handleOutOfMemoryError );
                try {
                    initMemoryReserve();
                } catch ( java.lang.OutOfMemoryError oome1 ) {
                    return grid2DSquareCell;
                }
            }
            throw oome0;
        }
    }
    
    /**
     * Returns a new Grid2DSquareCellDouble that is a copy of grid, but
     * with the value of cells in grid that intersect rectangle given by
     * (startRowIndex,startColIndex,endRowIndex,endColIndex) set to its
     * noDataValue.
     * @param grid The Grid2DSquareCellDouble to be masked.
     * @param startRowIndex The index of the first row to be masked.
     * @param startColIndex The index of the first column to be masked.
     * @param endRowIndex The index of the final row to be masked.
     * @param endColIndex The index of the final column to be masked.
     * @param grid2DSquareCellFactory The Grid2DSquareCellAbstractFactory used to
     *   generate result.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellAbstract mask( Grid2DSquareCellAbstract grid, long startRowIndex, long startColIndex, long endRowIndex, long endColIndex, Grid2DSquareCellAbstractFactory grid2DSquareCellFactory, boolean handleOutOfMemoryError ) {
        Grid2DSquareCellDouble result = ( Grid2DSquareCellDouble ) grid2DSquareCellFactory.create( grid );
        //result.setName( grid.getName() + "_mask" );
        return mask( result, startRowIndex, startColIndex, endRowIndex, endColIndex, this.handleOutOfMemoryErrorTrue );
    }
    
    /**
     * Returns grid with values linearly rescaled into the range [min,max].
     * @param grid the Grid2DSquareCellDouble to be processed.
     * @param min the minimum value in the rescaled range.
     * @param max the maximum value in the rescaled range.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDouble linearRescale( Grid2DSquareCellDouble grid, double min, double max, boolean handleOutOfMemoryError ) {
        long nrows = grid.getNrows(handleOutOfMemoryError);
        long ncols = grid.getNcols(handleOutOfMemoryError);
        double noDataValue = grid.getNoDataValue(handleOutOfMemoryError);
        double thisMin = grid.getGridStatistics().getMinDouble();
        double thisMax = grid.getGridStatistics().getMaxDouble();
        double thisValue;
        long row;
        long col;
        // if range of either input or output is zero
        if ( ( thisMax - thisMin ) == 0.0d || ( max - min ) == 0.0d ) {
            for ( row = 0; row < nrows; row ++ ) {
                for ( col = 0; col < ncols; col ++ ) {
                    thisValue = grid.getCell( row, col, this.handleOutOfMemoryErrorTrue );
                    if ( thisValue != noDataValue ) {
                        grid.setCell( row, col, min, this.handleOutOfMemoryErrorTrue );
                    }
                }
            }
        } else {
            for ( row = 0; row < nrows; row ++ ) {
                for ( col = 0; col < ncols; col ++ ) {
                    thisValue = grid.getCell( row, col, this.handleOutOfMemoryErrorTrue );
                    if ( thisValue != noDataValue ) {
                        grid.setCell( row, col, ( ( ( ( thisValue - thisMin ) / ( thisMax - thisMin ) ) ) * ( max - min ) ) + min, this.handleOutOfMemoryErrorTrue );
                    }
                }
            }
        }
        //grid.setName( grid.getName() + "_linearRescale" );
        return grid;
    }
    
    /**
     * Returns a new Grid2DSquareCellDouble that is a copy of grid, but
     * with values linearly rescaled into the range [min,max].
     * @param grid the Grid2DSquareCellDouble to be processed.
     * @param min the minimum value in the rescaled range.
     * @param max the maximum value in the rescaled range.
     * @param gridFactory the Grid2DSquareCellDoubleFactory used to
     *   generate result.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDouble linearRescale( Grid2DSquareCellDouble grid, double min, double max, Grid2DSquareCellDoubleFactory gridFactory, boolean handleOutOfMemoryError ) {
        Grid2DSquareCellDouble result = ( Grid2DSquareCellDouble ) gridFactory.create( grid );
        //result.setName( grid.getName() + "_linearRescale" );
        return linearRescale( result, min, max, this.handleOutOfMemoryErrorTrue );
    }
    
    /**
     * Returns grid with values linearly scaled to range [1.0d,1000000.0d] then
     * logged using Math.log(double) and then linearly scaled into the range
     * [min,max]
     * @param grid the Grid2DSquareCellDouble to be processed
     * @param min the minimum value in the rescaled range
     * @param max the maximum value in the rescaled range
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDouble logRescale( Grid2DSquareCellDouble grid, double min, double max, boolean handleOutOfMemoryError ) {
        long nrows = grid.getNrows(handleOutOfMemoryError);
        long ncols = grid.getNcols(handleOutOfMemoryError);
        double noDataValue = grid.getNoDataValue(handleOutOfMemoryError);
        double thisMin = grid.getGridStatistics().getMinDouble();
        double thisMax = grid.getGridStatistics().getMaxDouble();
        linearRescale( grid, 1.0d, 1000000.0d, this.handleOutOfMemoryErrorTrue );
        double thisValue;
        long row;
        long col;
        for ( row = 0; row < nrows; row ++ ) {
            for ( col = 0; col < ncols; col ++ ) {
                thisValue = grid.getCell( row, col, this.handleOutOfMemoryErrorTrue );
                if ( thisValue != noDataValue ) {
                    grid.setCell( row, col, Math.log( thisValue ), this.handleOutOfMemoryErrorTrue );
                }
            }
        }
        linearRescale( grid, min, max, this.handleOutOfMemoryErrorTrue );
        //grid.setName( grid.getName() + "_logRescale" );
        return grid;
    }
    
    /**
     * Returns a new Grid2DSquareCellDouble that is a copy of grid, but
     * with values linearly scaled to range [1.0d,1000000.0d] then logged using
     * Math.log(double) and then linearly scaled into the range [min,max]
     * @param grid the Grid2DSquareCellDouble to be processed
     * @param min the minimum value in the rescaled range
     * @param max the maximum value in the rescaled range
     * @param gridFactory the Grid2DSquareCellDoubleFactory used to
     *   generate result.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDouble logRescale( Grid2DSquareCellDouble grid, double min, double max, Grid2DSquareCellDoubleFactory gridFactory, boolean handleOutOfMemoryError ) {
        Grid2DSquareCellDouble result = ( Grid2DSquareCellDouble ) gridFactory.create( grid );
        //result.setName( grid.getName() + "_logRescale" );
        return logRescale( result, min, max, this.handleOutOfMemoryErrorTrue );
    }
    
    /**
     * Returns grid with value of cells with keys in cellKeys set to a value a
     * little bit larger.
     * @param grid the Grid2DSquareCellDouble to be processed
     * @param cellIDs
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDouble setValueALittleBitLarger( Grid2DSquareCellDouble grid, HashSet cellIDs, boolean handleOutOfMemoryError ) {
        CellID cellID;
        double noDataValue = grid.getNoDataValue(handleOutOfMemoryError);
        Iterator iterator1 = cellIDs.iterator();
        double thisValue;
        int counter = 0;
        while ( iterator1.hasNext() ) {
            cellID = ( ( CellID ) iterator1.next() );
            thisValue = grid.getCell( cellID.getCellRowIndex(), cellID.getCellColIndex(), this.handleOutOfMemoryErrorTrue );
            if ( thisValue != noDataValue ) {
                grid.setCell( cellID, Utilities.getValueALittleBitLarger( thisValue ), this.handleOutOfMemoryErrorTrue );
            }
        }
        return grid;
    }
    
    /**
     * Returns a new Grid2DSquareCellDouble that is a copy of grid, but
     * with value of cells with keys in cellKeys set to a value a little bit
     * larger.
     * @param grid the Grid2DSquareCellDouble to be processed
     * @param cellIDs
     * @param gridFactory the Grid2DSquareCellDoubleFactory used to
     *   generate result.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDouble setValueALittleBitLarger( Grid2DSquareCellDouble grid, HashSet cellIDs, Grid2DSquareCellDoubleFactory gridFactory, boolean handleOutOfMemoryError ) {
        Grid2DSquareCellDouble result = ( Grid2DSquareCellDouble ) gridFactory.create( grid );
        return setValueALittleBitLarger( result, cellIDs, this.handleOutOfMemoryErrorTrue );
    }
    
    /**
     * Returns grid with value of cells with keys in cellKeys set to a value a
     * little bit smaller.
     * @param grid the Grid2DSquareCellDouble to be processed
     * @param cellIDKeys
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDouble setValueALittleBitSmaller( Grid2DSquareCellDouble grid, HashSet cellIDKeys, boolean handleOutOfMemoryError ) {
        CellID cellID;
        double noDataValue = grid.getNoDataValue(handleOutOfMemoryError);
        Iterator iterator1 = cellIDKeys.iterator();
        double thisValue;
        while ( iterator1.hasNext() ) {
            cellID = ( CellID ) iterator1.next();
            thisValue = grid.getCell( cellID.getCellRowIndex(), cellID.getCellColIndex(), this.handleOutOfMemoryErrorTrue );
            if ( thisValue != noDataValue ) {
                grid.setCell( cellID, Utilities.getValueALittleBitSmaller( thisValue ), this.handleOutOfMemoryErrorTrue );
            }
        }
        return grid;
    }
    
    /**
     * Returns a new Grid2DSquareCellDouble that is a copy of grid, but
     * with value of cells with keys in cellKeys set to a value a little bit
     * smaller.
     * @param grid the Grid2DSquareCellDouble to be processed
     * @param cellIDs
     * @param gridFactory the Grid2DSquareCellDoubleFactory used to
     *   generate result.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDouble setValueALittleBitSmaller( Grid2DSquareCellDouble grid, HashSet cellIDs, Grid2DSquareCellDoubleFactory gridFactory, boolean handleOutOfMemoryError ) {
        Grid2DSquareCellDouble result = ( Grid2DSquareCellDouble ) gridFactory.create( grid );
        return setValueALittleBitSmaller( result, cellIDs, this.handleOutOfMemoryErrorTrue );
    }
    
    /**
     * Returns grid with value added to all cells with IDs as Integers in
     * cellKeys.
     * @param grid the Grid2DSquareCellDouble to be processed
     * @param cellIDs
     * @param value the value to be added
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDouble addToGrid( Grid2DSquareCellDouble grid, HashSet cellIDs, double value, boolean handleOutOfMemoryError ) {
        Iterator iterator1 = cellIDs.iterator();
        while ( iterator1.hasNext() ) {
            //grid.addToCell( ( CellID ) iterator1.next(), value );
            CellID cellID = ( CellID ) iterator1.next();
            if ( cellID != null ) {
                grid.addToCell( cellID, value, this.handleOutOfMemoryErrorTrue );
            }
        }
        return grid;
    }
    
    /**
     * Returns a copy of grid with value added to all cells with IDs as Integer
     * keys in cellIDKeys
     * @param grid the Grid2DSquareCellDouble to be processed
     * @param cellIDs
     * @param value the value to be added
     * @param gridFactory the Grid2DSquareCellDoubleFactory used to
     *   generate result.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDouble addToGrid( Grid2DSquareCellDouble grid, HashSet cellIDs, double value, Grid2DSquareCellDoubleFactory gridFactory, boolean handleOutOfMemoryError ) {
        Grid2DSquareCellDouble result = ( Grid2DSquareCellDouble ) gridFactory.create( grid );
        return addToGrid( result, cellIDs, value, this.handleOutOfMemoryErrorTrue );
    }
    
    /**
     * Returns grid with value added to every cell via
     * AbstractGrid2DSquareCellDouble.addToCell(value)
     * @param grid the Grid2DSquareCellDouble to be processed
     * @param value the value to be added
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDouble addToGrid( Grid2DSquareCellDouble grid, double value, boolean handleOutOfMemoryError ) {
        long nrows = grid.getNrows(handleOutOfMemoryError);
        long ncols = grid.getNcols(handleOutOfMemoryError);
        long row;
        long col;
        for ( row = 0; row < nrows; row ++ ) {
            for ( col = 0; col < ncols; col ++ ) {
                grid.addToCell( row, col, value, this.handleOutOfMemoryErrorTrue );
            }
        }
        return grid;
    }
    
    /**
     * Returns grid with value added to every cell via
     * Grid2DSquareCellDouble.addToCell(value)
     * @param grid the Grid2DSquareCellDouble to be processed
     * @param value the value to be added
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDouble addToGrid( Grid2DSquareCellDouble grid, CellID[] cellIDs, double value, boolean handleOutOfMemoryError ) {
        for ( int cellIDIndex = 0; cellIDIndex < cellIDs.length; cellIDIndex ++ ) {
            grid.addToCell( cellIDs[ cellIDIndex ].getCellRowIndex(), cellIDs[ cellIDIndex ].getCellColIndex(), value, this.handleOutOfMemoryErrorTrue );
        }
        return grid;
    }
    
    /**
     * Returns a copy of grid with value added to every cell via
     * AbstractGrid2DSquareCellDouble.addToCell(value)
     * @param grid the Grid2DSquareCellDouble to be processed
     * @param value the value to be added
     * @param gridFactory the Grid2DSquareCellDoubleFactory used to
     *   generate result.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDouble addToGrid( Grid2DSquareCellDouble grid, double value, Grid2DSquareCellDoubleFactory gridFactory, boolean handleOutOfMemoryError ) {
        Grid2DSquareCellDouble result = ( Grid2DSquareCellDouble ) gridFactory.create( grid );
        return addToGrid( result, value, this.handleOutOfMemoryErrorTrue );
    }
    
    /**
     * Returns grid with values added from gridToAdd
     * @param grid the Grid2DSquareCellDouble to be processed
     * @param gridToAdd the Grid2DSquareCellDouble from which values are
     *   added
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDouble addToGrid( Grid2DSquareCellDouble grid, Grid2DSquareCellDouble gridToAdd, boolean handleOutOfMemoryError ) {
        return addToGrid( grid, gridToAdd, 1.0d, this.handleOutOfMemoryErrorTrue );
    }
    
    /**
     * Returns a Grid2DSquareCellDouble with values of grid2DSquareCellDouble
     * added with values from grid2DSquareCellDoubleToAdd multiplied by weight.
     * @param grid2DSquareCellDouble The Grid2DSquareCellDouble to be processed.
     * @param grid2DSquareCellDoubleToAdd The Grid2DSquareCellDouble from which
     *   values are added.
     * @param weight The value grid2DSquareCellDoubleToAdd values are multiplied
     *   by.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDouble addToGrid( Grid2DSquareCellDouble grid2DSquareCellDouble, Grid2DSquareCellDouble grid2DSquareCellDoubleToAdd, double weight, boolean handleOutOfMemoryError ) {
        return addToGrid( grid2DSquareCellDouble, grid2DSquareCellDoubleToAdd, 0L, 0L, grid2DSquareCellDoubleToAdd.getNrows(handleOutOfMemoryError) - 1L, grid2DSquareCellDoubleToAdd.getNcols(handleOutOfMemoryError) - 1L, weight, this.handleOutOfMemoryErrorTrue );
    }
    
    /**
     * Returns a Grid2DSquareCellDouble with values of grid2DSquareCellDouble
     * added with values from grid2DSquareCellDoubleToAdd (with row index between
     * startRowIndex, endRowIndex and column index between startColIndex,
     * endColIndex) multiplied by weight.
     * @param grid2DSquareCellDouble The Grid2DSquareCellDouble to be processed.
     * @param grid2DSquareCellDoubleToAdd The Grid2DSquareCellDouble from which
     *   values are added.
     * @param startRowIndex The index of the first row from which
     *   grid2DSquareCellDoubleToAdd values are added.
     * @param startColIndex the index of the first column from which
     *   grid2DSquareCellDoubleToAdd values are added.
     * @param endRowIndex the index of the final row from which
     *   grid2DSquareCellDoubleToAdd values are added.
     * @param endColIndex the index of the final column from which
     *   grid2DSquareCellDoubleToAdd values are added.
     * @param weight The value grid2DSquareCellDoubleToAdd values are multiplied
     *   by.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDouble addToGrid( 
            Grid2DSquareCellDouble grid2DSquareCellDouble, 
            Grid2DSquareCellDouble grid2DSquareCellDoubleToAdd, 
            long startRowIndex, 
            long startColIndex, 
            long endRowIndex, 
            long endColIndex, 
            double weight, 
            boolean handleOutOfMemoryError ) {
        BigDecimal[] dimensions = grid2DSquareCellDoubleToAdd.getDimensions( handleOutOfMemoryError );
        BigDecimal[] dimensionConstraints = new BigDecimal[ 5 ];
        dimensionConstraints[ 1 ] = dimensions[ 1 ].add( new BigDecimal( startColIndex ).multiply( dimensions[ 0 ] ) );
        dimensionConstraints[ 2 ] = dimensions[ 2 ].add( new BigDecimal( startRowIndex ).multiply( dimensions[ 0 ] ) );
        dimensionConstraints[ 3 ] = dimensions[ 1 ].add( new BigDecimal( endColIndex - startColIndex + 1L ).multiply( dimensions[ 0 ] ) );
        dimensionConstraints[ 4 ] = dimensions[ 2 ].add( new BigDecimal( endRowIndex - startRowIndex + 1L ).multiply( dimensions[ 0 ] ) );
        return addToGrid( grid2DSquareCellDouble, grid2DSquareCellDoubleToAdd, startRowIndex, startColIndex, endRowIndex, endColIndex, dimensionConstraints, weight, this.handleOutOfMemoryErrorTrue );
    }
    
    //    /**
    //     * Returns grid with values added from gridToAdd for cells within bounds
    //     * given by (xmin,ymin,xmax,ymax) after they are multiplied by weight
    //     * @param grid2DSquareCellDouble The Grid2DSquareCellDouble to be processed.
    //     * @param grid2DSquareCellDoubleToAdd The Grid2DSquareCellDouble from which
    //     *   values are added.
    //     * @param dimensionContraints
    //     * @param weight The value grid2DSquareCellDoubleToAdd values are multiplied
    //     *   by.
    //     */
    //    public Grid2DSquareCellDouble addToGrid( Grid2DSquareCellDouble grid2DSquareCellDouble, Grid2DSquareCellDouble grid2DSquareCellDoubleToAdd, BigDecimal[] dimensionConstraints, double weight ) {
    //        //        long startRowIndex = gridToAdd.getCellRowIndex( ymin );
    //        //        long startColIndex = gridToAdd.getCellColIndex( xmin );
    //        //        long endRowIndex = gridToAdd.getCellRowIndex( ymax );
    //        //        long endColIndex = gridToAdd.getCellColIndex( xmax );
    //        BigDecimal cellsizeDividedBy2BigDecimal =
    //        long startRowIndex = grid2DSquareCellDoubleToAdd.getCellRowIndex( dimensionConstraints[ 2 ] );
    //        long startColIndex = grid2DSquareCellDoubleToAdd.getCellColIndex( dimensionConstraints[ 1 ] );
    //        long endRowIndex = grid2DSquareCellDoubleToAdd.getCellRowIndex( dimensionConstraints[ 4 ] );
    //        long endColIndex = grid2DSquareCellDoubleToAdd.getCellColIndex( dimensionConstraints[ 3 ] );
    //        return addToGrid( grid2DSquareCellDouble, grid2DSquareCellDoubleToAdd, startRowIndex, startColIndex, endRowIndex, endColIndex, dimensionConstraints, weight );
    //    }
    
    /**
     * Returns a Grid2DSquareCellDouble with values of grid2DSquareCellDouble
     * added with values from grid2DSquareCellDoubleToAdd (with row index between
     * startRowIndex, endRowIndex and column index between startColIndex,
     * endColIndex) multiplied by weight.
     * @param grid2DSquareCellDouble The Grid2DSquareCellDouble to be processed.
     * @param grid2DSquareCellDoubleToAdd The Grid2DSquareCellDouble from which
     *   values are added.
     * @param startRowIndex The index of the first row from which
     *   grid2DSquareCellDoubleToAdd values are added.
     * @param startColIndex the index of the first column from which
     *   grid2DSquareCellDoubleToAdd values are added.
     * @param endRowIndex the index of the final row from which
     *   grid2DSquareCellDoubleToAdd values are added.
     * @param endColIndex the index of the final column from which
     *   grid2DSquareCellDoubleToAdd values are added.
     * @param dimensionConstraints
     * @param weight The value grid2DSquareCellDoubleToAdd values are multiplied
     *   by.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * TODO: Check that reasonable answers are returned for intersections and aggregations.
     */
    protected Grid2DSquareCellDouble addToGrid( 
            Grid2DSquareCellDouble grid2DSquareCellDouble, 
            Grid2DSquareCellDouble grid2DSquareCellDoubleToAdd, 
            long startRowIndex, 
            long startColIndex, 
            long endRowIndex, 
            long endColIndex,
            BigDecimal[] dimensionConstraints, 
            double weight, 
            boolean handleOutOfMemoryError ) {
        long nrows = grid2DSquareCellDouble.getNrows(handleOutOfMemoryError);
        long ncols = grid2DSquareCellDouble.getNcols(handleOutOfMemoryError);
        double noDataValue = grid2DSquareCellDouble.getNoDataValue(handleOutOfMemoryError);
        BigDecimal[] grid2DSquareCellDoubleDimensions = grid2DSquareCellDouble.getDimensions( handleOutOfMemoryError);
        
        long grid2DSquareCellDoubleToAddNrows = grid2DSquareCellDoubleToAdd.getNrows(handleOutOfMemoryError);
        long grid2DSquareCellDoubleToAddNcols = grid2DSquareCellDoubleToAdd.getNcols(handleOutOfMemoryError);
        double grid2DSquareCellDoubleToAddNoDataValue = grid2DSquareCellDoubleToAdd.getNoDataValue(handleOutOfMemoryError);
        BigDecimal[] grid2DSquareCellDoubleToAddDimensions = grid2DSquareCellDoubleToAdd.getDimensions( handleOutOfMemoryError);
        
        //Grid2DSquareCellDoubleFactory grid2DSquareCellDoubleFactory = new Grid2DSquareCellDoubleFactory();
        Grid2DSquareCellDoubleFactory grid2DSquareCellDoubleFactory = new Grid2DSquareCellDoubleFactory( grid2DSquareCellDouble.getDirectory(handleOutOfMemoryError) );
        // TODO:
        // Implement higher method to pass in a Grid2DSquareCellDoubleChunkFactory.
        //Grid2DSquareCellDoubleFactory grid2DSquareCellDoubleFactory = new Grid2DSquareCellDoubleFactory( grid.getDirectory() );
        
        if ( ( dimensionConstraints[ 1 ].compareTo( grid2DSquareCellDoubleDimensions[ 3 ] ) == 1 ) ||
                ( dimensionConstraints[ 3 ].compareTo( grid2DSquareCellDoubleDimensions[ 1 ] ) == -1 ) ||
                ( dimensionConstraints[ 2 ].compareTo( grid2DSquareCellDoubleDimensions[ 4 ] ) == 1 ) ||
                ( dimensionConstraints[ 4 ].compareTo( grid2DSquareCellDoubleDimensions[ 2 ] ) == -1 ) ) {
            return grid2DSquareCellDouble;
        }
        if ( grid2DSquareCellDoubleToAddDimensions[ 0 ].compareTo( grid2DSquareCellDoubleDimensions[ 0 ] ) == -1 ) {
            // TODO:
            //            if ( startRowIndex != 0 || startColIndex != 0 || endRowIndex != grid2DSquareCellDoubleToAdd.getNrows() - 1 || endColIndex != grid2DSquareCellDoubleToAdd.getNcols() - 1 ) {
            //                System.out.println("Might be aggregating more than necessary!!!");
            //            }
            //            Grid2DSquareCellDouble aggregateGrid2DSquareCellDoubleToAddToAdd = null;
            //            try {
            //                aggregateGrid2DSquareCellDoubleToAddToAdd = aggregate( grid2DSquareCellDoubleToAdd, "sum", grid2DSquareCellDoubleDimensions, grid2DSquareCellDoubleFactory );
            //            } catch ( java.lang.OutOfMemoryError e ) {
            //                // TODO:
            //                // Clear some memory, cache, use chankRAF etc... try again. Perhaps fail and throw error to deal with higher up...
            //                System.out.println( "Warning!!! OutOfMemoryError in Grid2DSquareCellProcessor.addToGrid( grid ( ))" );
            //                int outOfMemoryErrorDebug0 = 1;
            //            }
            //            return addToGrid( grid2DSquareCellDouble, aggregateGrid2DSquareCellDoubleToAddToAdd, dimensionConstraints, weight );
            return null;
        } else {
            // If gridToAddCellsize is the same and the grids align
            if ( ( grid2DSquareCellDoubleToAddDimensions[ 0 ].compareTo( grid2DSquareCellDoubleDimensions[ 0 ] ) == 0 ) &&
                    ( ( grid2DSquareCellDoubleToAddDimensions[ 1 ].remainder( grid2DSquareCellDoubleDimensions[ 0 ] ) ).compareTo( ( grid2DSquareCellDoubleDimensions[ 1 ].remainder( grid2DSquareCellDoubleDimensions[ 0 ] ) ) ) == 0 ) &&
                    ( ( grid2DSquareCellDoubleToAddDimensions[ 2 ].remainder( grid2DSquareCellDoubleDimensions[ 0 ] ) ).compareTo( ( grid2DSquareCellDoubleDimensions[ 2 ].remainder( grid2DSquareCellDoubleDimensions[ 0 ] ) ) ) == 0 ) ) {
                //System.out.println( "Grids Align!" );
                double x;
                double y;
                double value;
                long row;
                long col;
                // TODO:
                // Make more robust if necessary using xBigDecimal and yBigDecimal
                // rather than using x and y?
                // The necessity can be calculated given the precision and size of
                // the double and the grid2DSquareCellDouble cellsize.
                for ( row = startRowIndex; row <= endRowIndex; row ++ ) {
                    for ( col = startColIndex; col <= endColIndex; col ++ ) {
                        x = grid2DSquareCellDoubleToAdd.getCellXDouble( col, this.handleOutOfMemoryErrorTrue );
                        y = grid2DSquareCellDoubleToAdd.getCellYDouble( row, this.handleOutOfMemoryErrorTrue );
                        value = grid2DSquareCellDoubleToAdd.getCell( row, col, this.handleOutOfMemoryErrorTrue );
                        if ( value != grid2DSquareCellDoubleToAddNoDataValue ) {
                            grid2DSquareCellDouble.addToCell( x, y, value * weight, this.handleOutOfMemoryErrorTrue );
                        }
                    }
                }
                return grid2DSquareCellDouble;
            } else {
                // System.out.println("Intersection!!!!");
                // Need to intersect
                // TODO:
                // Clip grid2DSquareCellDoubleToAdd might improve matters here.
                // Check
                Grid2DSquareCellDouble tempGrid1 = null;
                Grid2DSquareCellDouble tempGrid2 = null;
                try {
                    tempGrid1 = ( Grid2DSquareCellDouble ) grid2DSquareCellDoubleFactory.create( nrows, ncols, grid2DSquareCellDoubleDimensions );
                    tempGrid2 = ( Grid2DSquareCellDouble ) grid2DSquareCellDoubleFactory.create( nrows, ncols, grid2DSquareCellDoubleDimensions );
                } catch ( java.lang.OutOfMemoryError e ) {
                    System.out.println( "Warning!!! OutOfMemoryError: Unable to create temprorary grids trying another way but may run out of filespace!" );
                    // TODO:
                    // Clear some memory, cache, use chankRAF etc... try again. Perhaps fail and throw error to deal with higher up...
                    System.out.println( "Warning!!! OutOfMemoryError in Grid2DSquareCellProcessor.addToGrid( grid ( ))" );
                    int outOfMemoryErrorDebug1 = 1;
                }
                // TODO:
                // Check scale and rounding appropriate
                int scale = 324;
                int roundingMode = BigDecimal.ROUND_HALF_EVEN;
                BigDecimal grid2DSquareCellDoubleToAddCellsizeDividedBy2 = grid2DSquareCellDoubleToAddDimensions[ 0 ].divide( new BigDecimal( "2" ), scale, roundingMode );
                BigDecimal grid2DSquareCellDoubleToAddCellsizeSquared = grid2DSquareCellDoubleToAddDimensions[ 0 ].multiply( grid2DSquareCellDoubleToAddDimensions[ 0 ] );
                double[] bounds = new double[ 4 ];
                CellID cellID1;
                CellID cellID2;
                CellID cellID3;
                CellID cellID4;
                double d1;
                double d2;
                double d3;
                double d4;
                double x;
                double y;
                long row;
                long col;
                double areaProportion = 0.0d;
                // TODO:
                // precision checking and use of BigDecimal?
                for ( row = 0; row < nrows; row ++ ) {
                    for ( col = 0; col < ncols; col ++ ) {
                        bounds = grid2DSquareCellDouble.getCellBoundsDoubleArray( row, col, this.handleOutOfMemoryErrorTrue );
                        x = grid2DSquareCellDouble.getCellXDouble( col, this.handleOutOfMemoryErrorTrue );
                        y = grid2DSquareCellDouble.getCellYDouble( row, this.handleOutOfMemoryErrorTrue );
                        cellID1 = grid2DSquareCellDoubleToAdd.getCellID( bounds[ 0 ], bounds[ 3 ], this.handleOutOfMemoryErrorTrue );
                        cellID2 = grid2DSquareCellDoubleToAdd.getCellID( bounds[ 2 ], bounds[ 3 ], this.handleOutOfMemoryErrorTrue );
                        cellID3 = grid2DSquareCellDoubleToAdd.getCellID( bounds[ 0 ], bounds[ 1 ], this.handleOutOfMemoryErrorTrue );
                        cellID4 = grid2DSquareCellDoubleToAdd.getCellID( bounds[ 2 ], bounds[ 1 ], this.handleOutOfMemoryErrorTrue );
                        if ( cellID1.equals( cellID2 ) && cellID2.equals( cellID3 ) ) {
                            d1 = grid2DSquareCellDoubleToAdd.getCell( cellID1, this.handleOutOfMemoryErrorTrue );
                            if ( d1 != grid2DSquareCellDoubleToAddNoDataValue ) {
                                areaProportion = ( grid2DSquareCellDoubleDimensions[ 0 ].multiply( grid2DSquareCellDoubleDimensions[ 0 ] ).divide( grid2DSquareCellDoubleToAddCellsizeSquared, scale, roundingMode ) ).doubleValue();
                                tempGrid1.addToCell( row, col, d1 * areaProportion, this.handleOutOfMemoryErrorTrue );
                                tempGrid2.addToCell( row, col, areaProportion, this.handleOutOfMemoryErrorTrue );
                            }
                        } else {
                            d1 = grid2DSquareCellDoubleToAdd.getCell( cellID1, this.handleOutOfMemoryErrorTrue );
                            d2 = grid2DSquareCellDoubleToAdd.getCell( cellID2, this.handleOutOfMemoryErrorTrue );
                            d3 = grid2DSquareCellDoubleToAdd.getCell( cellID3, this.handleOutOfMemoryErrorTrue );
                            d4 = grid2DSquareCellDoubleToAdd.getCell( cellID4, this.handleOutOfMemoryErrorTrue );
                            
                            if ( ! grid2DSquareCellDoubleToAdd.isInGrid( cellID1, this.handleOutOfMemoryErrorTrue ) && d1 != grid2DSquareCellDoubleToAddNoDataValue ) {
                                if ( cellID1.equals( cellID2 ) || cellID1.equals( cellID3 ) ) {
                                    if ( cellID1.equals( cellID2 ) ) {
                                        //areaProportion = ( Math.abs( bounds[3] - ( grid2DSquareCellDoubleToAdd.getCellYDouble( cellID1 ) - grid2DSquareCellDoubleToAddCellsizeDividedBy2 ) ) * cellsize ) / ( grid2DSquareCellDoubleToAddCellsize * grid2DSquareCellDoubleToAddCellsize );
                                        areaProportion = Math.abs( ( ( BigDecimal.valueOf( bounds[3] ).subtract( grid2DSquareCellDoubleToAdd.getCellYBigDecimal( cellID1, this.handleOutOfMemoryErrorTrue ).subtract( grid2DSquareCellDoubleToAddCellsizeDividedBy2 ) ).multiply( grid2DSquareCellDoubleDimensions[ 0 ] ) ).divide( grid2DSquareCellDoubleToAddCellsizeSquared, scale, roundingMode ) ).doubleValue() );
                                    } else {
                                        //areaProportion = ( Math.abs( ( grid2DSquareCellDoubleToAdd.getCellXDouble( cellID1 ) + grid2DSquareCellDoubleToAddCellsizeDividedBy2 ) - bounds[0] ) * cellsize ) / ( grid2DSquareCellDoubleToAddCellsize * grid2DSquareCellDoubleToAddCellsize );
                                        areaProportion = Math.abs( ( ( ( ( grid2DSquareCellDoubleToAdd.getCellXBigDecimal( cellID1, this.handleOutOfMemoryErrorTrue ).add( grid2DSquareCellDoubleToAddCellsizeDividedBy2 ) ).subtract( BigDecimal.valueOf( bounds[0] ) ) ).multiply( grid2DSquareCellDoubleDimensions[ 0 ] ) ).divide( grid2DSquareCellDoubleToAddCellsizeSquared, scale, roundingMode ) ).doubleValue() );
                                    }
                                } else {
                                    //areaProportion = ( ( Math.abs( bounds[3] - ( grid2DSquareCellDoubleToAdd.getCellYDouble( cellID1 ) - grid2DSquareCellDoubleToAddCellsizeDividedBy2 ) ) * Math.abs( ( grid2DSquareCellDoubleToAdd.getCellXDouble( cellID1 ) + ( grid2DSquareCellDoubleToAddCellsize / 2.0d ) ) - bounds[0] ) ) / ( grid2DSquareCellDoubleToAddCellsize * grid2DSquareCellDoubleToAddCellsize ) );
                                    areaProportion = Math.abs( ( ( BigDecimal.valueOf( bounds[3] ).subtract( grid2DSquareCellDoubleToAdd.getCellYBigDecimal( cellID1, this.handleOutOfMemoryErrorTrue ).subtract( grid2DSquareCellDoubleToAddCellsizeDividedBy2 ) ) ).multiply( ( grid2DSquareCellDoubleToAdd.getCellXBigDecimal( cellID1, this.handleOutOfMemoryErrorTrue ).add( grid2DSquareCellDoubleToAddCellsizeDividedBy2.subtract( BigDecimal.valueOf( bounds[ 0 ] ) ) ) ).divide( grid2DSquareCellDoubleToAddCellsizeSquared, scale, roundingMode ) ) ).doubleValue() );
                                }
                                tempGrid1.addToCell( row, col, d1 * areaProportion, this.handleOutOfMemoryErrorTrue );
                                tempGrid2.addToCell( row, col, areaProportion, this.handleOutOfMemoryErrorTrue );
                            }
                            if ( ! grid2DSquareCellDoubleToAdd.isInGrid( cellID2, this.handleOutOfMemoryErrorTrue ) && d2 != grid2DSquareCellDoubleToAddNoDataValue ) {
                                if ( cellID2.equals( cellID1 ) ) {
                                    if ( cellID2.equals( cellID4 ) ) {
                                        //areaProportion = ( Math.abs( bounds[2] - ( grid2DSquareCellDoubleToAdd.getCellXDouble( cellID2 ) - grid2DSquareCellDoubleToAddCellsizeDividedBy2 ) ) * cellsize ) / ( grid2DSquareCellDoubleToAddCellsize * grid2DSquareCellDoubleToAddCellsize );
                                        areaProportion = Math.abs( ( ( ( BigDecimal.valueOf( bounds[ 2 ] ).subtract( grid2DSquareCellDoubleToAdd.getCellXBigDecimal( cellID2, this.handleOutOfMemoryErrorTrue ).subtract( grid2DSquareCellDoubleToAddCellsizeDividedBy2 ) ) ).multiply( grid2DSquareCellDoubleDimensions[ 0 ] ) ).divide( grid2DSquareCellDoubleToAddCellsizeSquared, scale, roundingMode ) ).doubleValue() );
                                    } else {
                                        //areaProportion = ( ( Math.abs( bounds[3] - ( grid2DSquareCellDoubleToAdd.getCellYDouble( cellID2 ) - grid2DSquareCellDoubleToAddCellsizeDividedBy2 ) ) * Math.abs( bounds[2] - ( grid2DSquareCellDoubleToAdd.getCellXDouble( cellID2 ) - ( grid2DSquareCellDoubleToAddCellsize / 2.0d ) ) ) ) / ( grid2DSquareCellDoubleToAddCellsize * grid2DSquareCellDoubleToAddCellsize ) );
                                        areaProportion = Math.abs( ( ( BigDecimal.valueOf( bounds[ 3 ] ).subtract( grid2DSquareCellDoubleToAdd.getCellYBigDecimal( cellID2, this.handleOutOfMemoryErrorTrue ).subtract( grid2DSquareCellDoubleToAddCellsizeDividedBy2 ) ) ).multiply( BigDecimal.valueOf( bounds[ 2 ] ).subtract( grid2DSquareCellDoubleToAdd.getCellXBigDecimal( cellID2, this.handleOutOfMemoryErrorTrue ).subtract( grid2DSquareCellDoubleToAddCellsizeDividedBy2 ) ) ).divide( grid2DSquareCellDoubleToAddCellsizeSquared, scale, roundingMode ) ).doubleValue() );
                                    }
                                    tempGrid1.addToCell( row, col, d2 * areaProportion, this.handleOutOfMemoryErrorTrue );
                                    tempGrid2.addToCell( row, col, areaProportion, this.handleOutOfMemoryErrorTrue );
                                }
                            }
                            if ( ! grid2DSquareCellDoubleToAdd.isInGrid( cellID3, this.handleOutOfMemoryErrorTrue ) && d3 != grid2DSquareCellDoubleToAddNoDataValue ) {
                                if ( ! cellID3.equals( cellID1 ) ) {
                                    if ( cellID3.equals( cellID4 ) ) {
                                        //areaProportion = ( Math.abs( ( grid2DSquareCellDoubleToAdd.getCellYDouble( cellID3 ) + ( grid2DSquareCellDoubleToAddCellsize / 2.0d ) ) - bounds[1] ) * cellsize ) / ( grid2DSquareCellDoubleToAddCellsize * grid2DSquareCellDoubleToAddCellsize );
                                        areaProportion = Math.abs( ( ( ( ( grid2DSquareCellDoubleToAdd.getCellYBigDecimal( cellID3, this.handleOutOfMemoryErrorTrue ).add( grid2DSquareCellDoubleToAddCellsizeDividedBy2 ) ).subtract( BigDecimal.valueOf( bounds[ 1 ] ) ) ).multiply( grid2DSquareCellDoubleDimensions[ 0 ] ) ).divide( grid2DSquareCellDoubleToAddCellsizeSquared, scale, roundingMode ) ).doubleValue() );
                                    } else {
                                        //areaProportion = ( ( Math.abs( ( grid2DSquareCellDoubleToAdd.getCellYDouble( cellID3 ) + ( grid2DSquareCellDoubleToAddCellsize / 2.0d ) ) - bounds[1] ) * Math.abs( ( grid2DSquareCellDoubleToAdd.getCellXDouble( cellID3 ) + ( grid2DSquareCellDoubleToAddCellsize / 2.0d ) ) - bounds[0] ) ) / ( grid2DSquareCellDoubleToAddCellsize * grid2DSquareCellDoubleToAddCellsize ) );
                                        areaProportion = Math.abs( ( ( ( ( grid2DSquareCellDoubleToAdd.getCellYBigDecimal( cellID3, this.handleOutOfMemoryErrorTrue ).add( grid2DSquareCellDoubleToAddCellsizeDividedBy2 ) ).subtract( BigDecimal.valueOf( bounds[ 1 ] ) ) ).multiply( ( grid2DSquareCellDoubleToAdd.getCellXBigDecimal( cellID3, this.handleOutOfMemoryErrorTrue ).add( grid2DSquareCellDoubleToAddCellsizeDividedBy2 ) ).subtract( BigDecimal.valueOf( bounds[ 0 ] ) ) ) ).divide( grid2DSquareCellDoubleToAddCellsizeSquared, scale, roundingMode ) ).doubleValue() );
                                    }
                                    tempGrid1.addToCell( row, col, d3 * areaProportion, this.handleOutOfMemoryErrorTrue );
                                    tempGrid2.addToCell( row, col, areaProportion, this.handleOutOfMemoryErrorTrue );
                                }
                            }
                            if ( ! grid2DSquareCellDoubleToAdd.isInGrid( cellID4, this.handleOutOfMemoryErrorTrue ) && d4 != grid2DSquareCellDoubleToAddNoDataValue ) {
                                if ( cellID4 != cellID2 && cellID4 != cellID3 ) {
                                    //areaProportion = ( ( Math.abs( ( grid2DSquareCellDoubleToAdd.getCellYDouble( cellID4 ) + ( grid2DSquareCellDoubleToAddCellsize / 2.0d ) ) - bounds[1] ) * Math.abs( bounds[2] - ( grid2DSquareCellDoubleToAdd.getCellXDouble( cellID4 ) - ( grid2DSquareCellDoubleToAddCellsize / 2.0d ) ) ) ) / ( grid2DSquareCellDoubleToAddCellsize * grid2DSquareCellDoubleToAddCellsize ) );
                                    areaProportion = Math.abs( ( ( ( ( grid2DSquareCellDoubleToAdd.getCellYBigDecimal( cellID4, this.handleOutOfMemoryErrorTrue ).add( grid2DSquareCellDoubleToAddCellsizeDividedBy2 ) ).subtract( BigDecimal.valueOf( bounds[ 1 ] ) ) ).multiply( BigDecimal.valueOf( bounds[ 2 ] ).subtract( ( grid2DSquareCellDoubleToAdd.getCellXBigDecimal( cellID4, this.handleOutOfMemoryErrorTrue ) ).subtract( grid2DSquareCellDoubleToAddCellsizeDividedBy2 ) ) ) ).divide( grid2DSquareCellDoubleToAddCellsizeSquared, scale, roundingMode ) ).doubleValue() );
                                    tempGrid1.addToCell( row, col, d4 * areaProportion, this.handleOutOfMemoryErrorTrue );
                                    tempGrid2.addToCell( row, col, areaProportion, this.handleOutOfMemoryErrorTrue );
                                }
                            }
                            //// Check
                            // Check fails due to rounding errors!
                            //if ( cellID1 != Integer.MIN_VALUE && cellID2 != Integer.MIN_VALUE && cellID3 != Integer.MIN_VALUE && cellID4 != Integer.MIN_VALUE && totalArea != 1.0 ) { System.out.println( "id = " + i + " : totalArea = " + totalArea + " (cellID1,cellID2,cellID3,cellID4) = (" + cellID1 + "," + cellID2 + "," + cellID3 + "," + cellID4 + ")" ); System.exit(0);}
                        }
                    }
                }
                System.out.println( "tempGrid1" );
                System.out.println( tempGrid1.toString() );
                System.out.println( "tempGrid2" );
                System.out.println( tempGrid2.toString() );
                // The values are normalised by dividing the aggregate Grid sum by the proportion of cells with grid values.
                for ( row = 0; row <= nrows; row ++ ) {
                    for ( col = 0; col <= ncols; col ++ ) {
                        d1 = tempGrid2.getCell( row, col, this.handleOutOfMemoryErrorTrue );
                        if ( ! ( d1 == 0.0d || d1 == noDataValue ) ) {
                            //setCell( i, tempGrid2.getCell( i ) );
                            grid2DSquareCellDouble.addToCell( row, col, weight * tempGrid1.getCell( row, col, this.handleOutOfMemoryErrorTrue ) / d1, this.handleOutOfMemoryErrorTrue );
                            //addToCell( i, tempGrid1.getCell( i ) / ( Math.pow( ( gridToAddCellsize / cellsize ), 2.0d ) / d1 ) );
                        }
                    }
                }
                //                tempGrid1.clear();
                //                tempGrid2.clear();
            }
        }
        return grid2DSquareCellDouble;
    }
    
    /**
     * Returns grid with values added from a file.
     * @param grid the Grid2DSquareCellDouble to be processed
     * @param file the file contining values to be added.
     * @param type the type of file.
     *   Supported types include "xyv", "xy", "idxy" )
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDouble addToGrid( Grid2DSquareCellDouble grid, File file, String type, boolean handleOutOfMemoryError ) {
        if ( type.equalsIgnoreCase( "xyv" ) ) {
            try {
                StreamTokenizer st =
                        new StreamTokenizer(
                        new BufferedReader(
                        new InputStreamReader(
                        new FileInputStream( file ) ) ) );
                st.eolIsSignificant( false );
                st.parseNumbers();
                st.whitespaceChars( ',', ',' );
                st.wordChars( '"', '"' );
                int tokenType = st.nextToken();
                String alternator = new String( "x" );
                double x = 0.0d;
                double y = 0.0d;
                double value = 0.0d;
                while ( tokenType != StreamTokenizer.TT_EOF ) {
                    switch ( tokenType ) {
                        case StreamTokenizer.TT_NUMBER :
                            if ( alternator.equals( "x" ) ) {
                                x = st.nval;
                                alternator = "y";
                            } else if ( alternator.equals( "y" ) ) {
                                y = st.nval;
                                alternator = "value";
                            } else {
                                grid.addToCell( x, y, st.nval, this.handleOutOfMemoryErrorTrue );
                                alternator = "x";
                            }
                            break;
                        default :
                            break;
                    }
                    tokenType = st.nextToken();
                }
            } catch ( java.io.IOException e ) {
                System.out.println( e.toString() );
                System.out.println( "in uk.ac.leeds.ccg.grids.AbstractGrid2DSquareCellDouble.addToGrid( file( " + file.toString() + "), type( " + type + " ) )");
                e.printStackTrace();
            }
        }
        if ( type.equalsIgnoreCase( "xy" ) ) {
            try {
                StreamTokenizer st = new StreamTokenizer( new BufferedReader( new InputStreamReader( new FileInputStream( file ) ) ) );
                st.eolIsSignificant( false );
                st.parseNumbers();
                st.whitespaceChars( ',', ',' );
                st.wordChars( '"', '"' );
                int tokenType = st.nextToken();
                String alternator = new String( "x" );
                double x = 0.0d;
                double y = 0.0d;
                double value = 0.0d;
                while ( tokenType != StreamTokenizer.TT_EOF ) {
                    switch ( tokenType ) {
                        case StreamTokenizer.TT_NUMBER :
                            if ( alternator.equals( "x" ) ) {
                                x = st.nval;
                                alternator = "y";
                            } else {
                                y = st.nval;
                                grid.addToCell( x, y, 1.0d, this.handleOutOfMemoryErrorTrue );
                                alternator = "x";
                            }
                            break;
                        default :
                            break;
                    }
                    tokenType = st.nextToken();
                }
            } catch ( java.io.IOException e ) {
                System.out.println( e.toString() );
                System.out.println( "in uk.ac.leeds.ccg.grids.AbstractGrid2DSquareCellDouble.addToGrid( file( " + file.toString() + "), type( " + type + " ) )");
                e.printStackTrace();
            }
        }
        if ( type.equalsIgnoreCase( "idxy" ) ) {
            try {
                StreamTokenizer st = new StreamTokenizer( new BufferedReader( new InputStreamReader( new FileInputStream( file ) ) ) );
                st.eolIsSignificant( false );
                st.parseNumbers();
                st.ordinaryChar( 'e' );
                st.ordinaryChar( 'd' );
                st.ordinaryChar( 'E' );
                st.ordinaryChar( 'D' );
                int tokenType = st.nextToken();
                int nextTokenType;
                String alternator = new String( "id" );
                int id;
                double x = 0.0d;
                double y = 0.0d;
                double value = 0.0d;
                while ( tokenType != StreamTokenizer.TT_EOF ) {
                    switch ( tokenType ) {
                        case StreamTokenizer.TT_NUMBER :
                            if ( alternator.equals( "id" ) ) {
                                id = ( int ) st.nval;
                                alternator = "x";
                            } else {
                                if ( alternator.equals( "x" ) ) {
                                    x = st.nval;
                                    nextTokenType = st.nextToken();
                                    if ( nextTokenType != StreamTokenizer.TT_NUMBER ) {
                                        nextTokenType = st.nextToken();
                                        nextTokenType = st.nextToken();
                                        x = x * Math.pow( 10.0, st.nval );
                                    } else {
                                        st.pushBack();
                                    }
                                    alternator = "y";
                                } else {
                                    y = st.nval;
                                    nextTokenType = st.nextToken();
                                    if ( nextTokenType != StreamTokenizer.TT_NUMBER ) {
                                        nextTokenType = st.nextToken();
                                        nextTokenType = st.nextToken();
                                        y = y * Math.pow( 10.0, st.nval );
                                    } else {
                                        st.pushBack();
                                    }
                                    alternator = "id";
                                    //System.out.println( " x, y = " + x + ", " + y );
                                    grid.addToCell( x, y, 1.0d, this.handleOutOfMemoryErrorTrue );
                                }
                            }
                            break;
                        default :
                            break;
                    }
                    tokenType = st.nextToken();
                }
            } catch ( java.io.IOException e ) {
                System.out.println( e.toString() );
                System.out.println( "in uk.ac.leeds.ccg.grids.AbstractGrid2DSquareCellDouble.addToGrid( file( " + file.toString() + "), type( " + type + " ) )");
                e.printStackTrace();
            }
        }
        return grid;
    }
    
    //    /**
    //     * TODO:
    //     * 1. Documentiation
    //     * 2. Change so Grid2DSquareCellDouble are Grid2DSquareCellAbstract
    //     * Returns a Grid2DSquareCellDouble that's values are those of grid0 minus
    //     * grid1.
    //     * @param data controls if result is produced even if one of the inputs is a
    //     *   NoData value
    //     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
    //     *   in this method then swap operations are initiated prior to retrying.
    //     *   If false then OutOfMemoryErrors are caught and thrown.
    //     * TODO:
    //     * see multiply first
    //     */
    //    public Grid2DSquareCellAbstract minus (
    //            Grid2DSquareCellAbstract grid0,
    //            Grid2DSquareCellAbstract grid1,
    //            boolean data,
    //            Grid2DSquareCellDoubleFactory gridFactory,
    //            boolean handleOutOfMemoryError ) {
    //        try {
    //            double grid0NoDataValue = grid0.getNoDataValue();
    //        double grid1NoDataValue = grid1.getNoDataValue();
    //        long nrows = grid0.getNrows();
    //        long ncols = grid0.getNcols();
    //        BigDecimal[] dimensions0 = grid0.getDimensions();
    //        Grid2DSquareCellDouble result = ( Grid2DSquareCellDouble ) gridFactory.create( grid0.getChunkNrows(this.handleOutOfMemoryErrorFalse), grid0.getChunkNcols(this.handleOutOfMemoryErrorFalse), nrows, ncols, dimensions0, grid0NoDataValue );
    //        double value0;
    //        double value1;
    //        long row;
    //        long col;
    //        if ( data ) {
    //            for ( row = 0; row < nrows; row ++ ) {
    //                for ( col = 0; col < ncols; col ++ ) {
    //                    value0 = grid0.getCell( row, col, this.handleOutOfMemoryErrorTrue );
    //                    if ( value0 == grid0NoDataValue ) {
    //                        value0 = 0.0d;
    //                        value1 = grid1.getCell( row, col, this.handleOutOfMemoryErrorTrue );
    //                        if ( value1 != grid1NoDataValue ) {
    //                            result.setCell( row, col, value0 - value1, this.handleOutOfMemoryErrorTrue );
    //                        }
    //                    } else {
    //                        value1 = grid1.getCell( row, col, this.handleOutOfMemoryErrorTrue );
    //                        if ( value1 == grid1NoDataValue ) {
    //                            value1 = 0.0d;
    //                        }
    //                        result.setCell( row, col, value0 - value1, this.handleOutOfMemoryErrorTrue );
    //                    }
    //                }
    //            }
    //        } else {
    //            for ( row = 0; row < nrows; row ++ ) {
    //                for ( col = 0; col < ncols; col ++ ) {
    //                    value0 = grid0.getCell( row, col, this.handleOutOfMemoryErrorTrue );
    //                    value1 = grid1.getCell( row, col, this.handleOutOfMemoryErrorTrue );
    //                    if ( value0 != grid0NoDataValue && value1 != grid1NoDataValue ) {
    //                        result.setCell( row, col, value0 - value1, this.handleOutOfMemoryErrorTrue );
    //                    }
    //                }
    //            }
    //        }
    //        //result.setName( "minus_" + grid0.getName( this.handleOutOfMemoryErrorTrue ) + "_" + grid1.getName( this.handleOutOfMemoryErrorTrue ) );
    //        return result;
    //    }
    //
    //    /**
    //     * TODO:
    //     * 1. Documentiation
    //     * 2. Change so Grid2DSquareCellDouble are Grid2DSquareCellAbstract
    //     * @param data controls if result is produced even if one of the inputs is a
    //     *   NoData value
    //     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
    //     *   in this method then swap operations are initiated prior to retrying.
    //     *   If false then OutOfMemoryErrors are caught and thrown.
    //     * @see Grid2DSquareCellProcessor#addToGrid(Grid2DSquareCellDouble,Grid2DSquareCellDouble,boolean)
    //     * TODO:
    //     * see multiply first
    //     */
    //    public Grid2DSquareCellDouble add( Grid2DSquareCellDouble grid0, Grid2DSquareCellDouble grid1, boolean data, Grid2DSquareCellDoubleFactory gridFactory, boolean handleOutOfMemoryError ) {
    //        double grid0NoDataValue = grid0.getNoDataValue();
    //        double grid1NoDataValue = grid1.getNoDataValue();
    //        long nrows = grid0.getNrows();
    //        long ncols = grid0.getNcols();
    //        BigDecimal[] dimensions0 = grid0.getDimensions();
    //        Grid2DSquareCellDouble result = gridFactory.create( grid0.getChunkNrows(this.handleOutOfMemoryErrorFalse), grid0.getChunkNcols(this.handleOutOfMemoryErrorFalse), nrows, ncols, dimensions0, grid0NoDataValue );
    //        long row;
    //        long col;
    //        double value0;
    //        double value1;
    //        if ( data ) {
    //            for ( row = 0; row < nrows; row ++ ) {
    //                for ( col = 0; col < ncols; col ++ ) {
    //                    value0 = grid0.getCell( row, col, this.handleOutOfMemoryErrorTrue );
    //                    if ( value0 == grid0NoDataValue ) {
    //                        value0 = 0.0d;
    //                        value1 = grid1.getCell( row, col, this.handleOutOfMemoryErrorTrue );
    //                        if ( value1 != grid1NoDataValue ) {
    //                            result.setCell( row, col, value0 + value1, this.handleOutOfMemoryErrorTrue );
    //                        }
    //                    } else {
    //                        value1 = grid1.getCell( row, col, this.handleOutOfMemoryErrorTrue );
    //                        if ( value1 == grid1NoDataValue ) {
    //                            value1 = 0.0d;
    //                        }
    //                        result.setCell( row, col, value0 + value1, this.handleOutOfMemoryErrorTrue );
    //                    }
    //                }
    //            }
    //        } else {
    //            for ( row = 0; row < nrows; row ++ ) {
    //                for ( col = 0; col < ncols; col ++ ) {
    //                    value0 = grid0.getCell( row, col, this.handleOutOfMemoryErrorTrue );
    //                    value1 = grid1.getCell( row, col, this.handleOutOfMemoryErrorTrue );
    //                    if ( value0 != grid0NoDataValue && value1 != grid1NoDataValue ) {
    //                        result.setCell( row, col, value0 + value1, this.handleOutOfMemoryErrorTrue );
    //                    }
    //                }
    //            }
    //        }
    //        //result.setName( "add_" + grid0.getName( this.handleOutOfMemoryErrorTrue ) + "_" + grid1.getName( this.handleOutOfMemoryErrorTrue ) );
    //        return result;
    //    }
    //
    //
    //    /**
    //     * Returns a Grid2DSquareCellAbstract which is grid2DSquareCell0 multiplied
    //     * by grid2DSquareCell1 or null if grid2DSquareCell0 and grid2DSquareCell1
    //     * are not coincident.
    //     * @param grid2DSquareCell0
    //     * @param grid2DSquareCell1
    //     * @param grid2DSquareCellAbstractFactory the Grid2DSquareCellAbstractFactory
    //     *   used to generate the result.
    //     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
    //     *   in this method then swap operations are initiated prior to retrying.
    //     *   If false then OutOfMemoryErrors are caught and thrown.
    //     * @TODO
    //     * Organise for chunk processing.
    //     */
    //    public Grid2DSquareCellAbstract multiply( Grid2DSquareCellAbstract grid2DSquareCell0, Grid2DSquareCellAbstract grid2DSquareCell1, Grid2DSquareCellAbstractFactory grid2DSquareCellAbstractFactory, boolean handleOutOfMemoryError ) {
    //        // Initialisation
    //        Grid2DSquareCellAbstract result;
    //        long nrows = grid2DSquareCell0.getNrows();
    //        long ncols = grid2DSquareCell0.getNcols();
    //        long row;
    //        long col;
    //
    //        if ( grid2DSquareCellAbstractFactory.getClass() == Grid2DSquareCellIntFactory.class ) {
    //            // Probably better off creating another way...
    //            //BigDecimal[] dimensions0 = grid2DSquareCell0.getDimensions();
    //            //result = grid2DSquareCellAbstractFactory.create( grid0.getChunkNrows(this.handleOutOfMemoryErrorFalse), grid0.getChunkNcols(this.handleOutOfMemoryErrorFalse), nrows, ncols, dimensions0, grid0NoDataValue );
    //            result = ( Grid2DSquareCellInt ) ( ( Grid2DSquareCellIntFactory ) grid2DSquareCellAbstractFactory ).create( grid2DSquareCell0 );
    //        } else {
    //            // ( grid2DSquareCellAbstractFactory.getClass() == Grid2DSquareCellDoubleFactory.class )
    //            result = ( Grid2DSquareCellDouble ) ( ( Grid2DSquareCellDoubleFactory ) grid2DSquareCellAbstractFactory ).create( grid2DSquareCell0 );
    //        }
    //        if ( grid2DSquareCell0.getClass() == Grid2DSquareCellInt.class ) {
    //            Grid2DSquareCellInt grid2DSquareCellInt0 = ( Grid2DSquareCellInt ) grid2DSquareCell0;
    //            int grid2DSquareCellInt0NoDataValue = grid2DSquareCellInt0.getNoDataValue( handleOutOfMemoryError );
    //            if ( grid2DSquareCell1.getClass() == Grid2DSquareCellInt.class ) {
    //                Grid2DSquareCellInt grid2DSquareCellInt1 = ( Grid2DSquareCellInt ) grid2DSquareCell1;
    //                int grid2DSquareCellInt1NoDataValue = grid2DSquareCellInt1.getNoDataValue( handleOutOfMemoryError );
    //                int value0;
    //                int value1;
    //                for ( row = 0; row < nrows; row ++ ) {
    //                    for ( col = 0; col < ncols; col ++ ) {
    //                        value0 = grid2DSquareCellInt0.getCell( row, col, this.handleOutOfMemoryErrorTrue );
    //                        value1 = grid2DSquareCellInt1.getCell( row, col, this.handleOutOfMemoryErrorTrue );
    //                        result.setCell( row, col, value0 * value1, this.handleOutOfMemoryErrorTrue );
    //                    }
    //                }
    //            } else {
    //                Grid2DSquareCellDouble grid2DSquareCellDouble1 = ( Grid2DSquareCellDouble ) grid2DSquareCell1;
    //                double grid2DSquareCellDouble1NoDataValue = grid2DSquareCellDouble1.getNoDataValue();
    //                int value0;
    //                double value1;
    //                for ( row = 0; row < nrows; row ++ ) {
    //                    for ( col = 0; col < ncols; col ++ ) {
    //                        value0 = grid2DSquareCellInt0.getCell( row, col, this.handleOutOfMemoryErrorTrue );
    //                        value1 = grid2DSquareCellDouble1.getCell( row, col, this.handleOutOfMemoryErrorTrue );
    //                        result.setCell( row, col, value0 * value1, this.handleOutOfMemoryErrorTrue );
    //                    }
    //                }
    //            }
    //        } else {
    //            // ( grid2DSquareCell0.getClass() == Grid2DSquareCellDouble.class )
    //            Grid2DSquareCellDouble grid2DSquareCellDouble0 = ( Grid2DSquareCellDouble ) grid2DSquareCell0;
    //            double grid2DSquareCellDouble0NoDataValue = grid2DSquareCellDouble0.getNoDataValue();
    //            if ( grid2DSquareCell1.getClass() == Grid2DSquareCellInt.class ) {
    //                Grid2DSquareCellInt grid2DSquareCellInt1 = ( Grid2DSquareCellInt ) grid2DSquareCell1;
    //                int grid2DSquareCellInt1NoDataValue = grid2DSquareCellInt1.getNoDataValue( handleOutOfMemoryError );
    //                double value0;
    //                int value1;
    //                for ( row = 0; row < nrows; row ++ ) {
    //                    for ( col = 0; col < ncols; col ++ ) {
    //                        value0 = grid2DSquareCellDouble0.getCell( row, col, this.handleOutOfMemoryErrorTrue );
    //                        value1 = grid2DSquareCellInt1.getCell( row, col, this.handleOutOfMemoryErrorTrue );
    //                        result.setCell( row, col, value0 * value1, this.handleOutOfMemoryErrorTrue );
    //                    }
    //                }
    //            } else {
    //                Grid2DSquareCellDouble grid2DSquareCellDouble1 = ( Grid2DSquareCellDouble ) grid2DSquareCell1;
    //                double grid2DSquareCellDouble1NoDataValue = grid2DSquareCellDouble1.getNoDataValue();
    //                double value0;
    //                double value1;
    //                for ( row = 0; row < nrows; row ++ ) {
    //                    for ( col = 0; col < ncols; col ++ ) {
    //                        value0 = grid2DSquareCellDouble0.getCell( row, col, this.handleOutOfMemoryErrorTrue );
    //                        value1 = grid2DSquareCellDouble1.getCell( row, col, this.handleOutOfMemoryErrorTrue );
    //                        result.setCell( row, col, value0 * value1, this.handleOutOfMemoryErrorTrue );
    //                    }
    //                }
    //            }
    //        }
    //        //result.setName( "multiply_" + grid0.getName( this.handleOutOfMemoryErrorTrue ) + "_" + grid1.getName( this.handleOutOfMemoryErrorTrue ), this.handleOutOfMemoryErrorTrue );
    //        return result;
    //    }
    //
    //    /**
    //     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
    //     *   in this method then swap operations are initiated prior to retrying.
    //     *   If false then OutOfMemoryErrors are caught and thrown.
    //     * TODO:
    //     * 1. Documentiation
    //     * 2. Change so Grid2DSquareCellDouble are Grid2DSquareCellAbstract
    //     */
    //    public Grid2DSquareCellDouble divide( Grid2DSquareCellDouble grid0, Grid2DSquareCellDouble grid1, Grid2DSquareCellDoubleFactory gridFactory, double defaultValue, boolean handleOutOfMemoryError ) {
    //        double grid0NoDataValue = grid0.getNoDataValue();
    //        double grid1NoDataValue = grid1.getNoDataValue();
    //        long nrows = grid0.getNrows();
    //        long ncols = grid0.getNcols();
    //        BigDecimal[] dimensions0 = grid0.getDimensions();
    //        Grid2DSquareCellDouble result = gridFactory.create( grid0.getChunkNrows(this.handleOutOfMemoryErrorFalse), grid0.getChunkNcols(this.handleOutOfMemoryErrorFalse), nrows, ncols, dimensions0, grid0NoDataValue );
    //        double value0;
    //        double value1;
    //        long row;
    //        long col;
    //        for ( row = 0; row < nrows; row ++ ) {
    //            for ( col = 0; col < ncols; col ++ ) {
    //                value0 = grid0.getCell( row, col, this.handleOutOfMemoryErrorTrue );
    //                value1 = grid1.getCell( row, col, this.handleOutOfMemoryErrorTrue );
    //                result.setCell( row, col, value0 / value1, this.handleOutOfMemoryErrorTrue );
    //            }
    //        }
    //        //result.setName( "divide_" + grid0.getName() + "_" + grid1.getName() );
    //        return result;
    //    }
    
    //    /**
    //     * Returns an Grid2DSquareCellDouble at a lower level of resolution than grid.  The result values
    //     * are either the sum, mean, max or min of values in grid depending on statistic.
    //     * xllcorner and yllcorner are the same as grid.
    //     * @param grid - the Grid2DSquareCellDouble to be processed
    //     * @param cellfactor - the number of times wider/higher the aggregated grid cells are
    //     * @param statistic - "sum", "mean", "max", or "min" depending on what aggregate of values are wanted
    //     * NB. In the calculation of the sum and the mean if there is a cell in grid which has a data value then
    //     *     the result which incorporates that cell has a data value.  For this result cell, any of the cells in
    //     *     grid which have noDataValues their value is taken as that of the average of its nearest cells with
    //     *     a value.
    //     *     In the calculation of the max and the min noDataValues are simply ignored.
    //     *     Formerly noDataValues were treated as the average of values within a result cell.
    //     * TODO:
    //     * 1. Implement median, mode and variance aggregations.
    //     */
    //    public Grid2DSquareCellDouble aggregate( Grid2DSquareCellDouble grid, int cellFactor, String statistic ) {
    //        try {
    //            return aggregate( grid, cellFactor, statistic, 0, 0, new Grid2DSquareCellDoubleJAIFactory() );
    //        } catch ( java.lang.OutOfMemoryError e ) {
    //            return aggregate( grid, cellFactor, statistic, 0, 0, new Grid2DSquareCellDoubleFileFactory() );
    //        }
    //    }
    
    /**
     * Returns an Grid2DSquareCellDouble at a lower level of resolution
     * than grid.  The result values are either the sum, mean, max or min of
     * values in grid depending on statistic.
     * @param grid2DSquareCell the Grid2DSquareCellDouble to be processed
     * @param cellFactor the number of times wider/higher the aggregated grid
     *                   cells are to be
     * @param statistic "sum", "mean", "max", or "min" depending on what
     *                   aggregate of values are wanted
     * @param rowOffset the number of rows above or below the origin of grid
     *                  where the aggregation is to start.
     *                  > 0 result yllcorner will be above grid yllcorner
     *                  < 0 result yllcorner will be below grid yllcorner
     * @param colOffset the number of columns above or below the origin of grid
     *                  where the aggregation is to start.
     *                  > 0 result xllcorner will be right of grid xllcorner
     *                  < 0 result xllcorner will be left of grid xllcorner
     * @param gridFactory the Abstract2DSquareCellDoubleFactory used to create
     *                    result and temporary AbstractGrid2DSquareCellDoubles.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * NB. In the calculation of the sum and the mean if there is a cell in grid which has a data value then
     *     the result which incorporates that cell has a data value.  For this result cell, any of the cells in
     *     grid which have noDataValues their value is taken as that of the average of its nearest cells with
     *     a value.
     *     In the calculation of the max and the min noDataValues are simply ignored.
     *     Formerly noDataValues were treated as the average of values within a result cell.
     * TODO:
     * implement median, mode and variance aggregations.
     */
    public Grid2DSquareCellDouble aggregate(
            Grid2DSquareCellAbstract grid2DSquareCell,
            int cellFactor,
            String statistic,
            int rowOffset,
            int colOffset,
            Grid2DSquareCellDoubleFactory gridFactory,
            boolean handleOutOfMemoryError ) {
        try {
            // Initial tests
            if ( cellFactor <= 0 ) {
                System.out.println( "Warning!!! cellFactor <= 0 : Returning null." );
                return null;
            }
            // Initialisation
            long nrows = grid2DSquareCell.getNrows(handleOutOfMemoryError);
            long ncols = grid2DSquareCell.getNcols(handleOutOfMemoryError);
            BigDecimal[] dimensions = grid2DSquareCell.getDimensions( handleOutOfMemoryError );
            double noDataValue = Double.NEGATIVE_INFINITY;
            if ( grid2DSquareCell.getClass() == Grid2DSquareCellInt.class ) {
                noDataValue = ( double ) ( ( Grid2DSquareCellInt ) grid2DSquareCell ).getNoDataValue( handleOutOfMemoryError );
            } else {
                if ( grid2DSquareCell.getClass() == Grid2DSquareCellDouble.class ) {
                    noDataValue = ( ( Grid2DSquareCellDouble ) grid2DSquareCell ).getNoDataValue(handleOutOfMemoryError);
                } else {
                    System.out.println( "Grid2DSquareCellAbstract not recognised in aggregate( Grid2DSquareCellAbstract( " + grid2DSquareCell.toString( handleOutOfMemoryError ) + ", cellFactor( " + cellFactor + " ), statistic( " + statistic + " ), rowOffset( " + rowOffset + " ), colOffset( " + colOffset + " ), gridFactory( " + gridFactory + " ),  handleOutOfMemoryError( " + handleOutOfMemoryError + " ) )" );
                    //throw new Exception();
                }
            }
            BigDecimal[] resultDimensions = new BigDecimal[ 5 ];
            resultDimensions[0] = dimensions[0].multiply( new BigDecimal( Integer.toString( cellFactor ) ) );
            resultDimensions[1] = dimensions[1].add( dimensions[0].multiply( new BigDecimal( Integer.toString( colOffset ) ) ) );
            resultDimensions[2] = dimensions[2].add( dimensions[0].multiply( new BigDecimal( Integer.toString( rowOffset ) ) ) );
            //double resultCellsize = cellsize * ( double ) cellFactor;
            //double width = cellsize * ncols;
            //double height = cellsize * nrows;
            //double resultXllcorner = xllcorner + ( colOffset * cellsize );
            //double resultYllcorner = yllcorner + ( rowOffset * cellsize );
            // Calculate resultNrows and resultHeight
            long resultNrows = 1L;
            BigDecimal resultHeight = new BigDecimal( resultDimensions[0].toString() );
            //double resultHeight = resultCellsize;
            while ( resultDimensions[ 2 ].add( resultHeight ).compareTo( dimensions[ 4 ] ) == -1 ) {
                resultNrows ++;
                resultHeight = resultHeight.add( resultDimensions[0] );
            }
            //while ( ( resultYllcorner + resultHeight ) < ( yllcorner + height ) ) {
            //    resultNrows ++;
            //    resultHeight += resultCellsize;
            //}
            // Calculate resultNcols and resultWidth
            long resultNcols = 1L;
            BigDecimal resultWidth = new BigDecimal( resultDimensions[0].toString() );
            //double resultWidth = resultCellsize;
            while ( resultDimensions[ 1 ].add( resultWidth ).compareTo( dimensions[ 3 ] ) == -1 ) {
                resultNrows ++;
                resultWidth = resultWidth.add( resultDimensions[0] );
            }
            //while ( ( resultXllcorner + resultWidth ) < ( xllcorner + width ) ) {
            //    resultNcols ++;
            //    resultWidth += resultCellsize;
            //}
            resultDimensions[3] = dimensions[1].add( resultWidth );
            resultDimensions[4] = dimensions[2].add( resultHeight );
            // Initialise result
            gridFactory.setNoDataValue( noDataValue );
            Grid2DSquareCellDouble result = ( Grid2DSquareCellDouble ) gridFactory.create( resultNrows, resultNcols, resultDimensions );
            
            long row;
            long col;
            double x;
            double y;
            double value;
            
            // sum
            if ( statistic.equalsIgnoreCase( "sum" ) ) {
                Grid2DSquareCellDouble count = ( Grid2DSquareCellDouble ) gridFactory.create( resultNrows, resultNcols, resultDimensions );
                Grid2DSquareCellDouble normaliser = ( Grid2DSquareCellDouble ) gridFactory.create( resultNrows, resultNcols, resultDimensions );
                for ( row = 0; row < nrows; row ++ ) {
                    for ( col = 0; col < ncols; col ++ ) {
                        x = grid2DSquareCell.getCellXDouble( col, this.handleOutOfMemoryErrorTrue );
                        y = grid2DSquareCell.getCellYDouble( row, this.handleOutOfMemoryErrorTrue );
                        if ( result.isInGrid( x, y, this.handleOutOfMemoryErrorTrue ) ) {
                            value = grid2DSquareCell.getCellDouble( row, col, this.handleOutOfMemoryErrorTrue );
                            if ( value != noDataValue ) {
                                count.addToCell( x, y, 1.0d, this.handleOutOfMemoryErrorTrue );
                                result.addToCell( x, y, value, this.handleOutOfMemoryErrorTrue );
                            }
                            normaliser.addToCell( x, y, 1.0d, this.handleOutOfMemoryErrorTrue );
                        }
                    }
                }
                //            // Add the nearest values for the noDataValues so long as there is a value
                //            for ( row = 0; row < nrows; row ++ ) {
                //                for ( col = 0; col < ncols; col ++ ) {
                //                    x = grid2DSquareCell.getCellXDouble( col, this.handleOutOfMemoryErrorTrue );
                //                    y = grid2DSquareCell.getCellYDouble( row, this.handleOutOfMemoryErrorTrue );
                //                    if ( result.inGrid( x, y, this.handleOutOfMemoryErrorTrue ) ) {
                //                        if ( dataCount.getCell( x, y, this.handleOutOfMemoryErrorTrue ) != noDataValue ) {
                //                            result.addToCell( x, y, grid2DSquareCell.getNearestValueDouble( row, col, this.handleOutOfMemoryErrorTrue ), this.handleOutOfMemoryErrorTrue );
                //                        }
                //                    }
                //                }
                //            }
                // Normalise
                double count0;
                for ( row = 0; row < resultNrows; row ++ ) {
                    for ( col = 0; col < resultNcols; col ++ ) {
                        count0 = count.getCell( row, col, this.handleOutOfMemoryErrorTrue );
                        if ( count0 != 0.0d ) {
                            result.setCell( row, col, ( ( result.getCell( row, col, this.handleOutOfMemoryErrorTrue ) * normaliser.getCell( row, col, this.handleOutOfMemoryErrorTrue ) ) / count0 ), this.handleOutOfMemoryErrorTrue );
                        }
                    }
                }
            }
            
            // mean
            if ( statistic.equalsIgnoreCase( "mean" ) ) {
                Grid2DSquareCellDouble numerator = ( Grid2DSquareCellDouble ) gridFactory.create( resultNrows, resultNcols, resultDimensions );
                Grid2DSquareCellDouble denominator = ( Grid2DSquareCellDouble ) gridFactory.create( resultNrows, resultNcols, resultDimensions );
                for ( row = 0; row < nrows; row ++ ) {
                    for ( col = 0; col < ncols; col ++ ) {
                        x = grid2DSquareCell.getCellXDouble( col, this.handleOutOfMemoryErrorTrue );
                        y = grid2DSquareCell.getCellYDouble( row, this.handleOutOfMemoryErrorTrue );
                        if ( result.isInGrid( x, y, this.handleOutOfMemoryErrorTrue ) ) {
                            value = grid2DSquareCell.getCellDouble( row, col, this.handleOutOfMemoryErrorTrue );
                            if ( value != noDataValue ) {
                                numerator.addToCell( x, y, value, this.handleOutOfMemoryErrorTrue );
                                denominator.addToCell( x, y, 1.0d, this.handleOutOfMemoryErrorTrue );
                            }
                        }
                    }
                }
                for ( row = 0; row < resultNrows; row ++ ) {
                    for ( col = 0; col < resultNcols; col ++ ) {
                        value = numerator.getCell( row, col, this.handleOutOfMemoryErrorTrue );
                        if ( value != noDataValue ) {
                            result.setCell( row, col, value / denominator.getCell( row, col, this.handleOutOfMemoryErrorTrue ), this.handleOutOfMemoryErrorTrue );
                        }
                    }
                }
            }
            
            // min
            if ( statistic.equalsIgnoreCase( "min" ) ) {
                double min;
                for ( row = 0; row < nrows; row ++ ) {
                    for ( col = 0; col < ncols; col ++ ) {
                        x = grid2DSquareCell.getCellXDouble( col, this.handleOutOfMemoryErrorTrue );
                        y = grid2DSquareCell.getCellYDouble( row, this.handleOutOfMemoryErrorTrue );
                        if ( result.isInGrid( x, y, this.handleOutOfMemoryErrorTrue ) ) {
                            value = grid2DSquareCell.getCellDouble( row, col, this.handleOutOfMemoryErrorTrue );
                            if ( value != noDataValue ) {
                                min = result.getCell( x, y, this.handleOutOfMemoryErrorTrue );
                                if ( min != noDataValue ) {
                                    result.setCell( x, y, Math.min( min, value ), this.handleOutOfMemoryErrorTrue );
                                } else {
                                    result.setCell( x, y, value, this.handleOutOfMemoryErrorTrue );
                                }
                            }
                        }
                    }
                }
            }
            
            // max
            if ( statistic.equalsIgnoreCase( "max" ) ) {
                double max;
                for ( row = 0; row < nrows; row ++ ) {
                    for ( col = 0; col < ncols; col ++ ) {
                        x = grid2DSquareCell.getCellXDouble( col, this.handleOutOfMemoryErrorTrue );
                        y = grid2DSquareCell.getCellYDouble( row, this.handleOutOfMemoryErrorTrue );
                        if ( result.isInGrid( x, y, this.handleOutOfMemoryErrorTrue ) ) {
                            value = grid2DSquareCell.getCellDouble( row, col, this.handleOutOfMemoryErrorTrue );
                            if ( value != noDataValue ) {
                                max = result.getCell( x, y, this.handleOutOfMemoryErrorTrue );
                                if ( max != noDataValue ) {
                                    result.setCell( x, y, Math.max( max, value ), this.handleOutOfMemoryErrorTrue );
                                } else {
                                    result.setCell( x, y, value, this.handleOutOfMemoryErrorTrue );
                                }
                            }
                        }
                    }
                }
            }
            return result;
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                Utilities.swapToFileChunks( this.grid2DSquareCells, handleOutOfMemoryError );
                return aggregate( grid2DSquareCell, cellFactor, statistic, rowOffset, colOffset, gridFactory, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
            
        }
    }
    
    //    /**
    //     * Returns an Grid2DSquareCellDouble at a lower level of resolution than grid.  The result values
    //     * are either the sum, mean, max or min of values in grid depending on statistic.
    //     * @param grid - the Grid2DSquareCellDouble to be processed
    //     * @param resultCellsize - output grid cellsize
    //     * @param statistic - "sum", "mean", "max", or "min" depending on what aggregate of values are wanted
    //     * @param resultXllcorner - the x-coordinate of the aggregate grid lower left corner
    //     * @param resultYllcorner - the y-coordinate of the aggregate grid lower left corner
    //     * Use this aggregate method to force origin of the result to be ( resultXllcorner, resultYllcorner ) and
    //     * if resultCellsize is not an integer multiple of this.cellsize.
    //     * NB. In the calculation of the sum and the mean if there is a cell in grid which has a data value then
    //     *     the result which incorporates that cell has a data value.  For this result cell, any of the cells in
    //     *     grid which have noDataValues their value is taken as that of the average of its nearest cells with
    //     *     a value.
    //     *     In the calculation of the max and the min noDataValues are simply ignored.
    //     *     Formerly noDataValues were treated as the average of values within a result cell.
    //     * TODO: implement median, mode and variance aggregations.
    //     */
    //    public Grid2DSquareCellDouble aggregate( Grid2DSquareCellDouble grid, double resultCellsize, String statistic, double resultXllcorner, double resultYllcorner ) {
    //        try {
    //            return aggregate( grid, resultCellsize, statistic, resultXllcorner, resultYllcorner, new Grid2DSquareCellDoubleFactory() );
    //        } catch ( java.lang.OutOfMemoryError e ) {
    //            return aggregate( grid, resultCellsize, statistic, resultXllcorner, resultYllcorner, new Grid2DSquareCellDoubleFileFactory() );
    //        }
    //    }
    
    /**
     * Returns an Grid2DSquareCellDouble at a lower level of resolution than grid.  The result values
     * are either the sum, mean, max or min of values in grid depending on statistic.
     * @param grid2DSquareCell The Grid2DSquareCellDouble to be processed
     * @param statistic "sum", "mean", "max", or "min" depending on what aggregate of values are wanted
     * @param resultDimensions
     * @param gridFactory The Abstract2DSquareCellDoubleFactory used to create grids
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * Use this aggregate method if result is to have a new spatial frame.
     * NB. In the calculation of the sum and the mean if there is a cell in grid which has a data value then
     *     the result which incorporates that cell has a data value.  For this result cell, any of the cells in
     *     grid which have noDataValues their value is taken as that of the average of its nearest cells with
     *     a value.
     *     In the calculation of the max and the min noDataValues are simply ignored.
     *     Formerly noDataValues were treated as the average of values within a result cell.
     * TODO:
     * implement median, mode and variance aggregations.
     */
    public Grid2DSquareCellDouble aggregate(
            Grid2DSquareCellAbstract grid2DSquareCell,
            String statistic,
            BigDecimal[] resultDimensions,
            Grid2DSquareCellDoubleFactory gridFactory,
            boolean handleOutOfMemoryError ) {
        int scale = 325;
        // Initialistaion
        long nrows = grid2DSquareCell.getNrows(handleOutOfMemoryError);
        long ncols = grid2DSquareCell.getNcols(handleOutOfMemoryError);
        BigDecimal[] dimensions = grid2DSquareCell.getDimensions( handleOutOfMemoryError );
        double noDataValue = Double.NEGATIVE_INFINITY;
        if ( grid2DSquareCell.getClass() == Grid2DSquareCellInt.class ) {
            noDataValue = ( double ) ( ( Grid2DSquareCellInt ) grid2DSquareCell ).getNoDataValue( handleOutOfMemoryError );
        } else {
            if ( grid2DSquareCell.getClass() == Grid2DSquareCellDouble.class ) {
                noDataValue = ( ( Grid2DSquareCellDouble ) grid2DSquareCell ).getNoDataValue(handleOutOfMemoryError);
            } else {
                System.out.println( "Grid2DSquareCellAbstract not recognised in aggregate( Grid2DSquareCellAbstract( " + grid2DSquareCell.toString( handleOutOfMemoryError ) + ", statistic( " + statistic + " ), resultDimensions( " + resultDimensions + " ), gridFactory( " + gridFactory + " ),  handleOutOfMemoryError( " + handleOutOfMemoryError + " ) )" );
                //throw new Exception();
            }
        }
        double cellsize = resultDimensions[0].doubleValue();
        
        //double width = cellsize * ncols;
        //double height = cellsize * nrows;
        
        // Test this is an aggregation
        if ( resultDimensions[0].compareTo( dimensions[ 0 ] ) != 1 ) {
            System.out.println( "!!!Warning: Not an aggregation as resultCellsize < cellsize. Returning null!" );
            return null;
        }
        // Test for intersection
        if ( ( resultDimensions[1].compareTo( dimensions[1].add( dimensions[ 0 ].multiply( new BigDecimal( Long.toString( ncols ) ) ) ) ) == 1 )
        || ( resultDimensions[2].compareTo( dimensions[2].add( dimensions[ 0 ].multiply( new BigDecimal( Long.toString( nrows ) ) ) ) ) == 1 ) ) {
            System.out.println( "!!!Warning: No intersection for aggregation. Returning null!" );
            return null;
        }
        // If resultCellsize is an integer multiple of cellsize and grid aligns with result then use
        // a cellFactor aggregation as it should be faster.
        //System.out.println("resultCellsize % cellsize == " + ( resultCellsize % cellsize ) );
        //System.out.println("resultXllcorner % cellsize = " + ( resultXllcorner % cellsize ) + ", xllcorner % cellsize = " + ( xllcorner % cellsize ) );
        //System.out.println("resultYllcorner % cellsize = " + ( resultYllcorner % cellsize ) + ", yllcorner % cellsize = " + ( yllcorner % cellsize ) );
        if ( true ) {
            BigDecimal t0 = resultDimensions[0].divide( dimensions[0], Math.max( resultDimensions[0].scale(), dimensions[0].scale() ) + 2, BigDecimal.ROUND_HALF_EVEN );
            BigDecimal t1 = resultDimensions[1].divide( dimensions[0], Math.max( resultDimensions[1].scale(), dimensions[0].scale() ) + 2, BigDecimal.ROUND_HALF_EVEN );
            BigDecimal t2 = dimensions[1].divide( dimensions[0], Math.max( dimensions[1].scale(), dimensions[0].scale() ) + 2, BigDecimal.ROUND_HALF_EVEN );
            BigDecimal t3 = resultDimensions[2].divide( dimensions[0], Math.max( resultDimensions[2].scale(), dimensions[0].scale() ) + 2, BigDecimal.ROUND_HALF_EVEN );
            BigDecimal t4 = dimensions[2].divide( dimensions[0], Math.max( dimensions[2].scale(), dimensions[0].scale() ) + 2, BigDecimal.ROUND_HALF_EVEN );
            if (
                    ( t0.compareTo( new BigDecimal( t0.toBigInteger().toString() ) ) == 0 ) &&
                    ( t1.compareTo( new BigDecimal( t1.toBigInteger().toString() ) ) == t2.compareTo( new BigDecimal( t2.toBigInteger().toString() ) ) ) &&
                    ( t3.compareTo( new BigDecimal( t3.toBigInteger().toString() ) ) == t4.compareTo( new BigDecimal( t4.toBigInteger().toString() ) ) ) ) {
                int cellFactor = resultDimensions[0].divide( dimensions[0], 2, BigDecimal.ROUND_UNNECESSARY ).intValue();
                int rowOffset = dimensions[2].subtract( resultDimensions[2].divide( dimensions[ 0 ], scale, BigDecimal.ROUND_HALF_EVEN ) ).intValue();
                int colOffset = dimensions[1].subtract( resultDimensions[1].divide( dimensions[ 0 ], scale, BigDecimal.ROUND_HALF_EVEN ) ).intValue();
                return aggregate( grid2DSquareCell, cellFactor, statistic, rowOffset, colOffset, gridFactory, this.handleOutOfMemoryErrorTrue );
            }
        }
        // Calculate resultNrows and resultHeight
        long resultNrows = 1L;
        BigDecimal resultHeight = new BigDecimal( resultDimensions[0].toString() );
        //double resultHeight = resultCellsize;
        while ( resultDimensions[ 2 ].add( resultHeight ).compareTo( dimensions[ 4 ] ) == -1 ) {
            resultNrows ++;
            resultHeight = resultHeight.add( resultDimensions[0] );
        }
        //while ( ( resultYllcorner + resultHeight ) < ( yllcorner + height ) ) {
        //    resultNrows ++;
        //    resultHeight += resultCellsize;
        //}
        // Calculate resultNcols and resultWidth
        long resultNcols = 1L;
        BigDecimal resultWidth = new BigDecimal( resultDimensions[0].toString() );
        //double resultWidth = resultCellsize;
        while ( resultDimensions[ 1 ].add( resultWidth ).compareTo( dimensions[ 3 ] ) == -1 ) {
            resultNrows ++;
            resultWidth = resultWidth.add( resultDimensions[0] );
        }
        //while ( ( resultXllcorner + resultWidth ) < ( xllcorner + width ) ) {
        //    resultNcols ++;
        //    resultWidth += resultCellsize;
        //}
        resultDimensions[3] = dimensions[1].add( resultWidth );
        resultDimensions[4] = dimensions[2].add( resultHeight );
        
        // Initialise result
        gridFactory.setNoDataValue( noDataValue );
        Grid2DSquareCellDouble result = ( Grid2DSquareCellDouble ) gridFactory.create( resultNrows, resultNcols, resultDimensions );
        
        long row;
        long col;
        double x;
        double y;
        double value;
        
        // sum
        if ( statistic.equalsIgnoreCase( "sum" ) ) {
            Grid2DSquareCellDouble totalValueArea = ( Grid2DSquareCellDouble ) gridFactory.create( resultNrows, resultNcols, resultDimensions );
            double areaProportion;
            double[] bounds = new double[ 4 ];
            CellID[] cellIDs = new CellID[ 4 ];
            
            double count0;
            for ( row = 0; row < nrows; row ++ ) {
                for ( col = 0; col < ncols; col ++ ) {
                    bounds = grid2DSquareCell.getCellBoundsDoubleArray( row, col, this.handleOutOfMemoryErrorTrue );
                    cellIDs[ 0 ] = result.getCellID( bounds[ 0 ], bounds[ 3 ], this.handleOutOfMemoryErrorTrue );
                    cellIDs[ 1 ] = result.getCellID( bounds[ 2 ], bounds[ 3 ], this.handleOutOfMemoryErrorTrue );
                    cellIDs[ 2 ] = result.getCellID( bounds[ 0 ], bounds[ 1 ], this.handleOutOfMemoryErrorTrue );
                    cellIDs[ 3 ] = result.getCellID( bounds[ 2 ], bounds[ 1 ], this.handleOutOfMemoryErrorTrue );
                    value = grid2DSquareCell.getCellDouble( row, col, this.handleOutOfMemoryErrorTrue );
                    if ( value != noDataValue ) {
                        if ( cellIDs[ 0 ].equals( cellIDs[ 1 ] ) && cellIDs[ 1 ].equals( cellIDs[2] ) ) {
                            result.addToCell( cellIDs[ 0 ], value, this.handleOutOfMemoryErrorTrue );
                            totalValueArea.addToCell( cellIDs[ 0 ], 1.0d, this.handleOutOfMemoryErrorTrue );
                        } else {
                            if ( cellIDs[ 0 ].equals( cellIDs[ 1 ] ) || cellIDs[ 0 ].equals( cellIDs[ 2 ] ) ) {
                                if ( cellIDs[ 0 ].equals( cellIDs[ 1 ] ) ) {
                                    areaProportion = ( Math.abs( bounds[3] - ( result.getCellYDouble( cellIDs[ 0 ], this.handleOutOfMemoryErrorTrue ) - ( resultDimensions[ 0 ].doubleValue() / 2.0d ) ) ) * cellsize ) / ( cellsize * cellsize );
                                } else {
                                    areaProportion = ( Math.abs( ( result.getCellXDouble( cellIDs[ 0 ], this.handleOutOfMemoryErrorTrue ) + ( resultDimensions[ 0 ].doubleValue() / 2.0d ) ) - bounds[0] ) * cellsize ) / ( cellsize * cellsize );
                                }
                            } else {
                                areaProportion = ( ( Math.abs( bounds[3] - ( result.getCellYDouble( cellIDs[ 0 ], this.handleOutOfMemoryErrorTrue ) - ( resultDimensions[ 0 ].doubleValue() / 2.0d ) ) ) * Math.abs( ( result.getCellXDouble( cellIDs[ 0 ], this.handleOutOfMemoryErrorTrue ) + ( resultDimensions[ 0 ].doubleValue() / 2.0d ) ) - bounds[0] ) ) / ( cellsize * cellsize ) );
                            }
                            result.addToCell( cellIDs[0], value * areaProportion, this.handleOutOfMemoryErrorTrue );
                            totalValueArea.addToCell( cellIDs[ 0 ], areaProportion, this.handleOutOfMemoryErrorTrue );
                        }
                        if ( ! cellIDs[ 1 ].equals( cellIDs[ 0 ] ) ) {
                            if ( cellIDs[ 1 ].equals( cellIDs[ 3 ] ) ) {
                                areaProportion = ( Math.abs( bounds[2] - ( result.getCellXDouble( cellIDs[ 1 ], this.handleOutOfMemoryErrorTrue ) - ( resultDimensions[ 0 ].doubleValue() / 2.0d ) ) ) * cellsize ) / ( cellsize * cellsize );
                            } else {
                                areaProportion = ( ( Math.abs( bounds[3] - ( result.getCellYDouble( cellIDs[ 1 ], this.handleOutOfMemoryErrorTrue ) - ( resultDimensions[ 0 ].doubleValue() / 2.0d ) ) ) * Math.abs( bounds[2] - ( result.getCellXDouble( cellIDs[ 1 ], this.handleOutOfMemoryErrorTrue ) - ( resultDimensions[ 0 ].doubleValue() / 2.0d ) ) ) ) / ( cellsize * cellsize ) );
                            }
                            result.addToCell( cellIDs[ 1 ], value * areaProportion, this.handleOutOfMemoryErrorTrue );
                            totalValueArea.addToCell( cellIDs[ 0 ], areaProportion, this.handleOutOfMemoryErrorTrue );
                        }
                        if ( ! cellIDs[ 2 ].equals( cellIDs[ 0 ] ) ) {
                            if ( ! cellIDs[ 2 ].equals( cellIDs[ 3 ] ) ) {
                                areaProportion = ( Math.abs( ( result.getCellYDouble( cellIDs[ 2 ], this.handleOutOfMemoryErrorTrue ) + ( resultDimensions[ 0 ].doubleValue() / 2.0d ) ) - bounds[1] ) * cellsize ) / ( cellsize * cellsize );
                            } else {
                                areaProportion = ( ( Math.abs( ( result.getCellYDouble( cellIDs[ 2 ], this.handleOutOfMemoryErrorTrue ) + ( resultDimensions[ 0 ].doubleValue() / 2.0d ) ) - bounds[1] ) * Math.abs( ( result.getCellXDouble( cellIDs[ 2 ], this.handleOutOfMemoryErrorTrue ) + ( resultDimensions[ 0 ].doubleValue() / 2.0d) ) - bounds[0] ) ) / ( cellsize * cellsize ) );
                            }
                            result.addToCell( cellIDs[ 2 ], value * areaProportion, this.handleOutOfMemoryErrorTrue );
                        }
                        if ( ! cellIDs[ 3 ].equals( cellIDs[ 1 ] ) && ! cellIDs[ 3 ].equals( cellIDs[ 2 ] ) ) {
                            areaProportion = ( ( Math.abs( ( result.getCellYDouble( cellIDs[ 3 ], this.handleOutOfMemoryErrorTrue ) + ( resultDimensions[ 0 ].doubleValue() / 2.0d ) ) - bounds[1] ) * Math.abs( bounds[2] - ( result.getCellXDouble( cellIDs[ 3 ], this.handleOutOfMemoryErrorTrue ) - ( resultDimensions[ 0 ].doubleValue() / 2.0d ) ) ) ) / ( cellsize * cellsize ) );
                            result.addToCell( cellIDs[ 3 ], value * areaProportion, this.handleOutOfMemoryErrorTrue );
                            totalValueArea.addToCell( cellIDs[ 0 ], areaProportion, this.handleOutOfMemoryErrorTrue );
                        }
                    }
                }
            }
            // Normalise
            double totalValueArea0;
            double resultCellsize = resultDimensions[ 0 ].doubleValue();
            for ( row = 0; row < resultNrows; row ++ ) {
                for ( col = 0; col < resultNcols; col ++ ) {
                    totalValueArea0 = totalValueArea.getCell( row, col, this.handleOutOfMemoryErrorTrue );
                    if ( totalValueArea0 != 0.0d ) {
                        result.setCell( row, col, ( ( result.getCell( row, col, this.handleOutOfMemoryErrorTrue ) * ( ( resultCellsize - cellsize ) * ( resultCellsize - cellsize ) ) ) / totalValueArea.getCell( row, col, this.handleOutOfMemoryErrorTrue ) ), this.handleOutOfMemoryErrorTrue );
                    }
                }
            }
            
        }
        //        // Add the nearest values for the noDataValues so long as there is a value
        //            for ( int i = 0; i < nrows; i ++ ) {
        //                for ( int j = 0; j < ncols; j ++ ) {
        //                    bounds = grid.getCellBounds( i, j );
        //                    cellID1 = result.getCellID( bounds[ 0 ], bounds[ 3 ] );
        //                    cellID2 = result.getCellID( bounds[ 2 ], bounds[ 3 ] );
        //                    cellID3 = result.getCellID( bounds[ 0 ], bounds[ 1 ] );
        //                    cellID4 = result.getCellID( bounds[ 2 ], bounds[ 1 ] );
        //                    if ( dataCount.getCell( bounds[ 0 ], bounds[ 3 ] ) != noDataValue ||
        //                    dataCount.getCell( bounds[ 2 ], bounds[ 3 ] ) != noDataValue ||
        //                    dataCount.getCell( bounds[ 0 ], bounds[ 1 ] ) != noDataValue ||
        //                    dataCount.getCell( bounds[ 2 ], bounds[ 1 ] ) != noDataValue ) {
        //                        value = grid.getNearestValueDouble( i, j );
        //                        if ( cellID1 == cellID2 && cellID2 == cellID3 ) {
        //                            if ( cellID1 != Integer.MIN_VALUE ) {
        //                                result.addToCell( cellID1, value );
        //                            }
        //                        } else {
        //                            if ( cellID1 != Integer.MIN_VALUE ) {
        //                                if ( cellID1 == cellID2 || cellID1 == cellID3 ) {
        //                                    if ( cellID1 == cellID2 ) {
        //                                        areaProportion = ( Math.abs( bounds[3] - ( result.getCellYDouble( cellID1 ) - ( resultCellsize / 2.0d ) ) ) * cellsize ) / ( cellsize * cellsize );
        //                                    } else {
        //                                        areaProportion = ( Math.abs( ( result.getCellXDouble( cellID1 ) + ( resultCellsize / 2.0d ) ) - bounds[0] ) * cellsize ) / ( cellsize * cellsize );
        //                                    }
        //                                } else {
        //                                    areaProportion = ( ( Math.abs( bounds[3] - ( result.getCellYDouble( cellID1 ) - ( resultCellsize / 2.0d ) ) ) * Math.abs( ( result.getCellXDouble( cellID1 ) + ( resultCellsize / 2.0d ) ) - bounds[0] ) ) / ( cellsize * cellsize ) );
        //                                }
        //                                result.addToCell( cellID1, value * areaProportion );
        //                            }
        //                            if ( cellID2 != Integer.MIN_VALUE ) {
        //                                if ( cellID2 != cellID1 ) {
        //                                    if ( cellID2 == cellID4 ) {
        //                                        areaProportion = ( Math.abs( bounds[2] - ( result.getCellXDouble( cellID2 ) - ( resultCellsize / 2.0d ) ) ) * cellsize ) / ( cellsize * cellsize );
        //                                    } else {
        //                                        areaProportion = ( ( Math.abs( bounds[3] - ( result.getCellYDouble( cellID2 ) - ( resultCellsize / 2.0d ) ) ) * Math.abs( bounds[2] - ( result.getCellXDouble( cellID2 ) - ( resultCellsize / 2.0d ) ) ) ) / ( cellsize * cellsize ) );
        //                                    }
        //                                    result.addToCell( cellID2, value * areaProportion );
        //                                }
        //                            }
        //                            if ( cellID3 != Integer.MIN_VALUE ) {
        //                                if ( cellID3 != cellID1 ) {
        //                                    if ( cellID3 == cellID4 ) {
        //                                        areaProportion = ( Math.abs( ( result.getCellYDouble( cellID3 ) + ( resultCellsize / 2.0d ) ) - bounds[1] ) * cellsize ) / ( cellsize * cellsize );
        //                                    } else {
        //                                        areaProportion = ( ( Math.abs( ( result.getCellYDouble( cellID3 ) + ( resultCellsize / 2.0d ) ) - bounds[1] ) * Math.abs( ( result.getCellXDouble( cellID3 ) + ( resultCellsize / 2.0d) ) - bounds[0] ) ) / ( cellsize * cellsize ) );
        //                                    }
        //                                    result.addToCell( cellID3, value * areaProportion );
        //                                }
        //                            }
        //                            if ( cellID4 != Integer.MIN_VALUE ) {
        //                                if ( cellID4 != cellID2 && cellID4 != cellID3 ) {
        //                                    areaProportion = ( ( Math.abs( ( result.getCellYDouble( cellID4 ) + ( resultCellsize / 2.0d ) ) - bounds[1] ) * Math.abs( bounds[2] - ( result.getCellXDouble( cellID4 ) - ( resultCellsize / 2.0d ) ) ) ) / ( cellsize * cellsize ) );
        //                                    result.addToCell( cellID4, value * areaProportion );
        //                                }
        //                            }
        //                        }
        //                    }
        //                }
        //            }
        //        }
        
        // mean
        if ( statistic.equalsIgnoreCase( "mean" ) ) {
            double denominator = ( resultDimensions[ 0 ].doubleValue() * resultDimensions[ 0 ].doubleValue() ) / ( cellsize * cellsize );
            Grid2DSquareCellDouble sum = aggregate( grid2DSquareCell, "sum", resultDimensions, gridFactory, this.handleOutOfMemoryErrorTrue);
            addToGrid( result, sum, 1.0d / denominator, this.handleOutOfMemoryErrorTrue );
        }
        
        // max
        if ( statistic.equalsIgnoreCase( "max" ) ) {
            double max;
            double[] bounds = new double[4];
            for ( row = 0; row < nrows; row ++ ) {
                for ( col = 0; col < ncols; col ++ ) {
                    value = grid2DSquareCell.getCellDouble( row, col, this.handleOutOfMemoryErrorTrue );
                    if ( value != noDataValue ) {
                        x = grid2DSquareCell.getCellXDouble( col, this.handleOutOfMemoryErrorTrue );
                        y = grid2DSquareCell.getCellYDouble( row, this.handleOutOfMemoryErrorTrue );
                        bounds = grid2DSquareCell.getCellBoundsDoubleArray( row, col, this.handleOutOfMemoryErrorTrue );
                        max = result.getCell( bounds[0], bounds[3], this.handleOutOfMemoryErrorTrue );
                        if ( max != noDataValue ) {
                            result.setCell( bounds[0], bounds[3], Math.max( max, value ), this.handleOutOfMemoryErrorTrue );
                        } else {
                            result.setCell( bounds[0], bounds[3], value, this.handleOutOfMemoryErrorTrue );
                        }
                        max = result.getCell( bounds[2], bounds[3], this.handleOutOfMemoryErrorTrue );
                        if ( max != noDataValue ) {
                            result.setCell( bounds[2], bounds[3], Math.max( max, value ), this.handleOutOfMemoryErrorTrue );
                        } else {
                            result.setCell( bounds[2], bounds[3], value, this.handleOutOfMemoryErrorTrue );
                        }
                        max = result.getCell( bounds[0], bounds[1], this.handleOutOfMemoryErrorTrue );
                        if ( max != noDataValue ) {
                            result.setCell( bounds[0], bounds[1], Math.max( max, value ), this.handleOutOfMemoryErrorTrue );
                        } else {
                            result.setCell( bounds[0], bounds[1], value, this.handleOutOfMemoryErrorTrue );
                        }
                        max = result.getCell( bounds[2], bounds[1], this.handleOutOfMemoryErrorTrue );
                        if ( max != noDataValue ) {
                            result.setCell( bounds[2], bounds[1], Math.max( max, value ), this.handleOutOfMemoryErrorTrue );
                        } else {
                            result.setCell( bounds[2], bounds[1], value, this.handleOutOfMemoryErrorTrue );
                        }
                    }
                }
            }
        }
        
        // min
        if ( statistic.equalsIgnoreCase( "min" ) ) {
            double min;
            double[] bounds = new double[4];
            for ( row = 0; row < nrows; row ++ ) {
                for ( col = 0; col < ncols; col ++ ) {
                    value = grid2DSquareCell.getCellDouble( row, col, this.handleOutOfMemoryErrorTrue );
                    if ( value != noDataValue ) {
                        x = grid2DSquareCell.getCellXDouble( col, this.handleOutOfMemoryErrorTrue );
                        y = grid2DSquareCell.getCellYDouble( row, this.handleOutOfMemoryErrorTrue );
                        bounds = grid2DSquareCell.getCellBoundsDoubleArray( row, col, this.handleOutOfMemoryErrorTrue );
                        min = result.getCell( bounds[0], bounds[3], this.handleOutOfMemoryErrorTrue );
                        if ( min != noDataValue ) {
                            result.setCell( bounds[0], bounds[3], Math.min( min, value ), this.handleOutOfMemoryErrorTrue );
                        } else {
                            result.setCell( bounds[0], bounds[3], value, this.handleOutOfMemoryErrorTrue );
                        }
                        min = result.getCell( bounds[2], bounds[3], this.handleOutOfMemoryErrorTrue );
                        if ( min != noDataValue ) {
                            result.setCell( bounds[2], bounds[3], Math.min( min, value ), this.handleOutOfMemoryErrorTrue );
                        } else {
                            result.setCell( bounds[2], bounds[3], value, this.handleOutOfMemoryErrorTrue );
                        }
                        min = result.getCell( bounds[0], bounds[1], this.handleOutOfMemoryErrorTrue );
                        if ( min != noDataValue ) {
                            result.setCell( bounds[0], bounds[1], Math.min( min, value ), this.handleOutOfMemoryErrorTrue );
                        } else {
                            result.setCell( bounds[0], bounds[1], value, this.handleOutOfMemoryErrorTrue );
                        }
                        min = result.getCell( bounds[2], bounds[1], this.handleOutOfMemoryErrorTrue );
                        if ( min != noDataValue ) {
                            result.setCell( bounds[2], bounds[1], Math.min( min, value ), this.handleOutOfMemoryErrorTrue );
                        } else {
                            result.setCell( bounds[2], bounds[1], value, this.handleOutOfMemoryErrorTrue );
                        }
                    }
                }
            }
        }
        
        /*
        // Initialistaion
        int nrows = grid.getNrows();
        int ncols = grid.getNcols();
        double xllcorner = grid.getXllcorner();
        double yllcorner = grid.getYllcorner();
        double cellsize = grid.getCellsize();
        double noDataValue = grid.getNoDataValue();
        double width = cellsize * ncols;
        double height = cellsize * nrows;
        // Test this is an aggregation
        if ( resultCellsize <= cellsize ) {
            System.out.println( "!!!Warning: resultCellsize <= cellsize in aggregate( cellsize( " + resultCellsize + " ), statistic( " + statistic + " ), resultXllcorner( " + resultXllcorner + " ), resultYllcorner( " + resultYllcorner + " ), noDataValue( " + noDataValue + " ), gridFactory( " + gridFactory.toString() + " ) ). Returning null!" );
            return null;
        }
        // Test for intersection
        if ( ( resultXllcorner > xllcorner + ( ( double ) ncols * cellsize ) ) || ( resultYllcorner > yllcorner + ( ( double ) nrows * cellsize ) ) ) {
            System.out.println( "!!!Warning: No intersection in aggregate( cellsize( " + resultCellsize + " ), statistic( " + statistic + " ), resultXllcorner( " + resultXllcorner + " ), resultYllcorner( " + resultYllcorner + " ), noDataValue( " + noDataValue + " ), gridFactory( " + gridFactory.toString() + " ) ). Returning null!" );
            return null;
        }
        // If resultCellsize is an integer multiple of cellsize and grid aligns with result then use
        // a cellFactor aggregation as it should be faster.
        //System.out.println("resultCellsize % cellsize == " + ( resultCellsize % cellsize ) );
        //System.out.println("resultXllcorner % cellsize = " + ( resultXllcorner % cellsize ) + ", xllcorner % cellsize = " + ( xllcorner % cellsize ) );
        //System.out.println("resultYllcorner % cellsize = " + ( resultYllcorner % cellsize ) + ", yllcorner % cellsize = " + ( yllcorner % cellsize ) );
        if ( ( resultCellsize % cellsize == 0.0d ) && ( ( resultXllcorner % cellsize ) == ( xllcorner % cellsize ) ) && ( ( resultYllcorner % cellsize ) == ( yllcorner % cellsize ) ) ) {
            int cellFactor = ( int ) ( resultCellsize / cellsize );
            int rowOffset = ( int ) ( yllcorner - resultYllcorner / cellsize );
            int colOffset = ( int ) ( xllcorner - resultXllcorner / cellsize );
            return aggregate( grid, cellFactor, statistic, rowOffset, colOffset, gridFactory );
        }
        // Calculate resultNrows and resultHeight
        int resultNrows = 1;
        double resultHeight = resultCellsize;
        while ( ( resultYllcorner + resultHeight ) < ( yllcorner + height ) ) {
            resultNrows ++;
            resultHeight += resultCellsize;
        }
        // Calculate resultNcols and resultWidth
        int resultNcols = 1;
        double resultWidth = resultCellsize;
        while ( ( resultXllcorner + resultWidth ) < ( xllcorner + width ) ) {
            resultNcols ++;
            resultWidth += resultCellsize;
        }
        //System.out.println( "resultNcols " + resultNcols + ", resultNrows " + resultNrows );
        //System.out.println( "gridToAddNcols " + ncols + ", gridToAddNrows " + nrows );
        // Initialise result
        Grid2DSquareCellDouble result = gridFactory.createGrid2DSquareCellDouble( resultNrows, resultNcols, resultXllcorner, resultYllcorner, resultCellsize, noDataValue );
         
        // sum
        if ( statistic.equalsIgnoreCase( "sum" ) ) {
            Grid2DSquareCellDouble tempGrid1 = gridFactory.createGrid2DSquareCellDouble( resultNrows, resultNcols, resultXllcorner, resultYllcorner, resultCellsize, noDataValue, 1 );
            Grid2DSquareCellDouble tempGrid2 = gridFactory.createGrid2DSquareCellDouble( resultNrows, resultNcols, resultXllcorner, resultYllcorner, resultCellsize, noDataValue, 1 );
            double x = 0.0d;
            double y = 0.0d;
            double d1 = noDataValue;
            double areaProportion = 0.0d;
            double[] bounds = new double[ 4 ];
            int cellID1 = -1;
            int cellID2 = -1;
            int cellID3 = -1;
            int cellID4 = -1;
            //double totalArea = 0;
            for ( int i = 0; i < nrows * ncols; i ++ ) {
                d1 = grid.getCell( i );
                if ( d1 != noDataValue ) {
                    x = grid.getCellXDouble( i );
                    y = grid.getCellYDouble( i );
                    bounds = grid.getCellBounds( i );
                    cellID1 = result.getCellID( bounds[ 0 ], bounds[ 3 ] );
                    cellID2 = result.getCellID( bounds[ 2 ], bounds[ 3 ] );
                    cellID3 = result.getCellID( bounds[ 0 ], bounds[ 1 ] );
                    cellID4 = result.getCellID( bounds[ 2 ], bounds[ 1 ] );
                    if ( cellID1 == cellID2 && cellID2 == cellID3 ) {
                        if ( cellID1 != Integer.MIN_VALUE ) {
                            areaProportion = 1.0d;
                            tempGrid1.addToCell( x, y, d1 * areaProportion );
                            tempGrid2.addToCell( x, y, areaProportion );
                        }
                    } else {
                        if ( cellID1 != Integer.MIN_VALUE ) {
                            if ( cellID1 == cellID2 || cellID1 == cellID3 ) {
                                if ( cellID1 == cellID2 ) {
                                    areaProportion = ( Math.abs( bounds[3] - ( result.getCellYDouble( cellID1 ) - ( resultCellsize / 2.0d ) ) ) * cellsize ) / ( cellsize * cellsize );
                                } else {
                                    areaProportion = ( Math.abs( ( result.getCellXDouble( cellID1 ) + ( resultCellsize / 2.0d ) ) - bounds[0] ) * cellsize ) / ( cellsize * cellsize );
                                }
                            } else {
                                areaProportion = ( ( Math.abs( bounds[3] - ( result.getCellYDouble( cellID1 ) - ( resultCellsize / 2.0d ) ) ) * Math.abs( ( result.getCellXDouble( cellID1 ) + ( resultCellsize / 2.0d ) ) - bounds[0] ) ) / ( cellsize * cellsize ) );
                            }
                            tempGrid1.addToCell( cellID1, d1 * areaProportion );
                            tempGrid2.addToCell( cellID1, areaProportion );
                        }
                        if ( cellID2 != Integer.MIN_VALUE ) {
                            if ( cellID2 != cellID1 ) {
                                if ( cellID2 == cellID4 ) {
                                    areaProportion = ( Math.abs( bounds[2] - ( result.getCellXDouble( cellID2 ) - ( resultCellsize / 2.0d ) ) ) * cellsize ) / ( cellsize * cellsize );
                                } else {
                                    areaProportion = ( ( Math.abs( bounds[3] - ( result.getCellYDouble( cellID2 ) - ( resultCellsize / 2.0d ) ) ) * Math.abs( bounds[2] - ( result.getCellXDouble( cellID2 ) - ( resultCellsize / 2.0d ) ) ) ) / ( cellsize * cellsize ) );
                                }
                                tempGrid1.addToCell( cellID2, d1 * areaProportion );
                                tempGrid2.addToCell( cellID2, areaProportion );
                            }
                        }
                        if ( cellID3 != Integer.MIN_VALUE ) {
                            if ( cellID3 != cellID1 ) {
                                if ( cellID3 == cellID4 ) {
                                    areaProportion = ( Math.abs( ( result.getCellYDouble( cellID3 ) + ( resultCellsize / 2.0d ) ) - bounds[1] ) * cellsize ) / ( cellsize * cellsize );
                                } else {
                                    areaProportion = ( ( Math.abs( ( result.getCellYDouble( cellID3 ) + ( resultCellsize / 2.0d ) ) - bounds[1] ) * Math.abs( ( result.getCellXDouble( cellID3 ) + ( resultCellsize / 2.0d) ) - bounds[0] ) ) / ( cellsize * cellsize ) );
                                }
                                tempGrid1.addToCell( cellID3, d1 * areaProportion );
                                tempGrid2.addToCell( cellID3, areaProportion );
                            }
                        }
                        if ( cellID4 != Integer.MIN_VALUE ) {
                            if ( cellID4 != cellID2 && cellID4 != cellID3 ) {
                                areaProportion = ( ( Math.abs( ( result.getCellYDouble( cellID4 ) + ( resultCellsize / 2.0d ) ) - bounds[1] ) * Math.abs( bounds[2] - ( result.getCellXDouble( cellID4 ) - ( resultCellsize / 2.0d ) ) ) ) / ( cellsize * cellsize ) );
                                tempGrid1.addToCell( cellID4, d1 * areaProportion );
                                tempGrid2.addToCell( cellID4, areaProportion );
                            }
                        }
                        // Check fails due to rounding errors!
                        //if ( cellID1 != Integer.MIN_VALUE && cellID2 != Integer.MIN_VALUE && cellID3 != Integer.MIN_VALUE && cellID4 != Integer.MIN_VALUE && totalArea != 1.0 ) { System.out.println( "id = " + i + " : totalArea = " + totalArea + " (cellID1,cellID2,cellID3,cellID4) = (" + cellID1 + "," + cellID2 + "," + cellID3 + "," + cellID4 + ")" );
                        //    throw an exception!!!
                        //}
                    }
                }
            }
            // The values are normalised by dividing the aggregate Grid sum by the proportion of cells with grid values.
            for ( int i = 0; i < resultNrows * resultNcols; i ++ ) {
                d1 = tempGrid2.getCell( i );
                if ( d1 != 0.0d && d1 != noDataValue ) {
                    result.setCell( i, tempGrid1.getCell( i ) / ( Math.pow( ( resultCellsize / cellsize ), 2.0d ) / d1 ) );
                }
            }
            tempGrid1.clear();
            tempGrid2.clear();
        }
         
        // mean
        if ( statistic.equalsIgnoreCase( "mean" ) ) {
            // To calculate the mean and cope with NODATA it is necessary to pass
            // through the data twice or for each aggregated cell get all
            // intersecting cells. This is because each cells area as a proportion
            // of the non noDataValue area of an aggregated cell is needed. This
            // cannot be simply done as it all depends on NODATA. In the
            // implementation below the data is read through twice. First read
            // involves calculating NODATA in each aggregated cell.
            Grid2DSquareCellDouble tempGrid1 = gridFactory.createGrid2DSquareCellDouble( resultNrows, resultNcols, resultXllcorner, resultYllcorner, resultCellsize, noDataValue, 1 );
            Grid2DSquareCellDouble tempGrid2 = gridFactory.createGrid2DSquareCellDouble( resultNrows, resultNcols, resultXllcorner, resultYllcorner, resultCellsize, noDataValue, 1 );
            double x;
            double y;
            double d1;
            double area;
            double[] bounds = new double[4];
            int cellID1;
            int cellID2;
            int cellID3;
            int cellID4;
            //double totalArea = 0;
            for ( int i = 0; i < nrows * ncols; i ++ ) {
                d1 = grid.getCell( i );
                if ( d1 != noDataValue ) {
                    x = grid.getCellXDouble( i );
                    y = grid.getCellYDouble( i );
                    bounds = grid.getCellBounds( i );
                    cellID1 = result.getCellID( bounds[0], bounds[3] );
                    cellID2 = result.getCellID( bounds[2], bounds[3] );
                    cellID3 = result.getCellID( bounds[0], bounds[1] );
                    cellID4 = result.getCellID( bounds[2], bounds[1] );
                    if ( cellID1 == cellID2 && cellID2 == cellID3 ) {
                        if ( cellID1 != Integer.MIN_VALUE ) {
                            area = cellsize * cellsize;
                            tempGrid1.addToCell( x, y, area );
                        }
                    } else {
                        if ( cellID1 != Integer.MIN_VALUE ) {
                            if ( cellID1 == cellID2 || cellID1 == cellID3 ) {
                                if ( cellID1 == cellID2 ) {
                                    area = ( Math.abs( bounds[3] - ( result.getCellYDouble( cellID1 ) - ( resultCellsize / 2.0d ) ) ) * cellsize );
                                } else {
                                    area = ( Math.abs( ( result.getCellXDouble( cellID1 ) + ( resultCellsize / 2.0d ) ) - bounds[0] ) * cellsize );
                                }
                            } else {
                                area = ( Math.abs( bounds[3] - ( result.getCellYDouble( cellID1 ) - ( resultCellsize / 2.0d ) ) ) * Math.abs( ( result.getCellXDouble( cellID1 ) + ( resultCellsize / 2.0d ) ) - bounds[0] ) );
                            }
                            tempGrid1.addToCell( cellID1, area );
                        }
                        if ( cellID2 != Integer.MIN_VALUE ) {
                            if ( cellID2 != cellID1 ) {
                                if ( cellID2 == cellID4 ) {
                                    area = ( Math.abs( bounds[2] - ( result.getCellXDouble( cellID2 ) - ( resultCellsize / 2.0d ) ) ) * cellsize );
                                } else {
                                    area = ( Math.abs( bounds[3] - ( result.getCellYDouble( cellID2 ) - ( resultCellsize / 2.0d ) ) ) * Math.abs( bounds[2] - ( result.getCellXDouble( cellID2 ) - ( resultCellsize / 2.0d ) ) ) );
                                }
                                tempGrid1.addToCell( cellID2, area );
                            }
                        }
                        if ( cellID3 != Integer.MIN_VALUE ) {
                            if ( cellID3 != cellID1 ) {
                                if ( cellID3 == cellID4 ) {
                                    area = ( Math.abs( ( result.getCellYDouble( cellID3 ) + ( resultCellsize / 2.0d ) ) - bounds[1] ) * cellsize );
                                } else {
                                    area = ( Math.abs( ( result.getCellYDouble( cellID3 ) + ( resultCellsize / 2.0d ) ) - bounds[1] ) * Math.abs( ( result.getCellXDouble( cellID3 ) + ( resultCellsize / 2.0d ) ) - bounds[0] ) );
                                }
                                tempGrid1.addToCell( cellID3, area );
                            }
                        }
                        if ( cellID4 != Integer.MIN_VALUE ) {
                            if ( cellID4 != cellID2 && cellID4 != cellID3 ) {
                                area = ( Math.abs( ( result.getCellYDouble( cellID4 ) + ( resultCellsize / 2.0d ) ) - bounds[1] ) * Math.abs( bounds[2] - ( result.getCellXDouble( cellID4 ) - ( resultCellsize / 2.0d ) ) ) );
                                tempGrid1.addToCell( cellID4, area );
                            }
                        }
                    }
                }
            }
            for ( int i = 0; i < nrows * ncols; i ++ ) {
                double areaIntersect;
                d1 = grid.getCell( i );
                if ( d1 != noDataValue ) {
                    x = grid.getCellXDouble( i );
                    y = grid.getCellYDouble( i );
                    bounds = grid.getCellBounds( i );
                    cellID1 = result.getCellID( bounds[0], bounds[3] );
                    cellID2 = result.getCellID( bounds[2], bounds[3] );
                    cellID3 = result.getCellID( bounds[0], bounds[1] );
                    cellID4 = result.getCellID( bounds[2], bounds[1] );
                    if ( cellID1 == cellID2 && cellID2 == cellID3 ) {
                        if ( cellID1 != Integer.MIN_VALUE ) {
                            area = tempGrid1.getCell( x, y );
                            result.addToCell( x, y, d1 * ( cellsize * cellsize ) / area );
                        }
                    } else {
                        if ( cellID1 != Integer.MIN_VALUE ) {
                            if ( cellID1 == cellID2 || cellID1 == cellID3 ) {
                                if ( cellID1 == cellID2 ) {
                                    areaIntersect = ( Math.abs( bounds[3] - ( result.getCellYDouble( cellID1 ) - ( resultCellsize / 2.0d ) ) ) * cellsize );
                                } else {
                                    areaIntersect = ( Math.abs( ( result.getCellXDouble( cellID1 ) + ( resultCellsize / 2.0d ) ) - bounds[0] ) * cellsize );
                                }
                            } else {
                                areaIntersect = ( Math.abs( bounds[3] - ( result.getCellYDouble( cellID1 ) - ( resultCellsize / 2.0d ) ) ) * Math.abs( ( result.getCellXDouble( cellID1 ) + ( resultCellsize / 2.0d ) ) - bounds[0] ) );
                            }
                            area = tempGrid1.getCell( cellID1 );
                            result.addToCell( cellID1, d1 * areaIntersect / area );
                        }
                        if ( cellID2 != Integer.MIN_VALUE ) {
                            if ( cellID2 != cellID1 ) {
                                if ( cellID2 == cellID4 ) {
                                    areaIntersect = ( Math.abs( bounds[2] - ( result.getCellXDouble( cellID2 ) - ( resultCellsize / 2.0d ) ) ) * cellsize );
                                } else {
                                    areaIntersect = ( Math.abs( bounds[3] - ( result.getCellYDouble( cellID2 ) - ( resultCellsize / 2.0d ) ) ) * Math.abs( bounds[2] - ( result.getCellXDouble( cellID2 ) - ( resultCellsize / 2.0d ) ) ) );
                                }
                                area = tempGrid1.getCell( cellID2 );
                                result.addToCell( cellID2, d1 * areaIntersect / area );
                            }
                        }
                        if ( cellID3 != Integer.MIN_VALUE ) {
                            if ( cellID3 != cellID1 ) {
                                if ( cellID3 == cellID4 ) {
                                    areaIntersect = ( Math.abs( ( result.getCellYDouble( cellID3 ) + ( resultCellsize / 2.0d ) ) - bounds[1] ) * cellsize );
                                } else {
                                    areaIntersect = ( Math.abs( ( result.getCellYDouble( cellID3 ) + ( resultCellsize / 2.0d ) ) - bounds[1] ) * Math.abs( ( result.getCellXDouble( cellID3 ) + ( resultCellsize / 2.0d ) ) - bounds[0] ) );
                                }
                                area = tempGrid1.getCell( cellID3 );
                                result.addToCell( cellID3, d1 * areaIntersect / area );
                            }
                        }
                        if ( cellID4 != Integer.MIN_VALUE ) {
                            if ( cellID4 != cellID2 && cellID4 != cellID3 ) {
                                areaIntersect = ( Math.abs( ( result.getCellYDouble( cellID4 ) + ( resultCellsize / 2.0d ) ) - bounds[1] ) * Math.abs( bounds[2] - ( result.getCellXDouble( cellID4 ) - ( resultCellsize / 2.0d ) ) ) );
                                area = tempGrid1.getCell( cellID4 );
                                result.addToCell( cellID4, d1 * areaIntersect / area );
                            }
                        }
                    }
                }
            }
            tempGrid1.clear();
        }
         
        // max
        if ( statistic.equalsIgnoreCase( "max" ) ) {
            double x;
            double y;
            double d1;
            double d2;
            double[] bounds = new double[4];
            for ( int i = 0; i < nrows * ncols; i ++ ) {
                d1 = grid.getCell( i );
                if ( d1 != noDataValue ) {
                    x = grid.getCellXDouble( i );
                    y = grid.getCellYDouble( i );
                    bounds = grid.getCellBounds( i );
                    d2 = result.getCell( bounds[0], bounds[3] );
                    if ( d2 != noDataValue ) {
                        result.setCell( bounds[0], bounds[3], Math.max( d2, d1 ) );
                    } else {
                        result.setCell( bounds[0], bounds[3], d1 );
                    }
                    d2 = result.getCell( bounds[2], bounds[3] );
                    if ( d2 != noDataValue ) {
                        result.setCell( bounds[2], bounds[3], Math.max( d2, d1 ) );
                    } else {
                        result.setCell( bounds[2], bounds[3], d1 );
                    }
                    d2 = result.getCell( bounds[0], bounds[1] );
                    if ( d2 != noDataValue ) {
                        result.setCell( bounds[0], bounds[1], Math.max( d2, d1 ) );
                    } else {
                        result.setCell( bounds[0], bounds[1], d1 );
                    }
                    d2 = result.getCell( bounds[2], bounds[1] );
                    if ( d2 != noDataValue ) {
                        result.setCell( bounds[2], bounds[1], Math.max( d2, d1 ) );
                    } else {
                        result.setCell( bounds[2], bounds[1], d1 );
                    }
                }
            }
        }
         
        // min
        if ( statistic.equalsIgnoreCase( "min" ) ) {
            double x;
            double y;
            double d1;
            double d2;
            double[] bounds = new double[4];
            for ( int i = 0; i < nrows * ncols; i ++ ) {
                d1 = grid.getCell( i );
                if ( d1 != noDataValue ) {
                    x = grid.getCellXDouble( i );
                    y = grid.getCellYDouble( i );
                    bounds = grid.getCellBounds( i );
                    d2 = result.getCell( bounds[0], bounds[3] );
                    if ( d2 != noDataValue ) {
                        result.setCell( bounds[0], bounds[3], Math.min( d2, d1 ) );
                    } else {
                        result.setCell( bounds[0], bounds[3], d1 );
                    }
                    d2 = result.getCell( bounds[2], bounds[3] );
                    if ( d2 != noDataValue ) {
                        result.setCell( bounds[2], bounds[3], Math.min( d2, d1 ) );
                    } else {
                        result.setCell( bounds[2], bounds[3], d1 );
                    }
                    d2 = result.getCell( bounds[0], bounds[1] );
                    if ( d2 != noDataValue ) {
                        result.setCell( bounds[0], bounds[1], Math.min( d2, d1 ) );
                    } else {
                        result.setCell( bounds[0], bounds[1], d1 );
                    }
                    d2 = result.getCell( bounds[2], bounds[1] );
                    if ( d2 != noDataValue ) {
                        result.setCell( bounds[2], bounds[1], Math.min( d2, d1 ) );
                    } else {
                        result.setCell( bounds[2], bounds[1], d1 );
                    }
                }
            }
        }
         */
        return result;
    }
    
    //    TODO: Move to a Extended Statistics class
    //    /**
    //     * Returns a double global statistic for an Grid2DSquareCellDouble grid
    //     * NB. Only for grids in the same spatial frame.
    //     */
    //    public double globalBivariateStatistics( Grid2DSquareCellDouble grid0, Grid2DSquareCellDouble grid1, String comparator  ) {
    //        // Initialisation
    //        int grid0Nrows = grid0.getNrows();
    //        int grid0Ncols = grid0.getNcols();
    //        double grid0Xllcorner = grid0.getXllcorner();
    //        double grid0Yllcorner = grid0.getYllcorner();
    //        double grid0Cellsize = grid0.getCellsize();
    //        double grid0NoDataValue = grid0.getNoDataValue();
    //        int grid1Nrows = grid1.getNrows();
    //        int grid1Ncols = grid1.getNcols();
    //        double grid1Xllcorner = grid1.getXllcorner();
    //        double grid1Yllcorner = grid1.getYllcorner();
    //        double grid1Cellsize = grid1.getCellsize();
    //        double grid1NoDataValue = grid1.getNoDataValue();
    //        AbstractGridStatistics grid0Statistics = grid0.getGridStatistics();
    //        AbstractGridStatistics grid1Statistics = grid1.getGridStatistics();
    //        // TODO: Check spatial frame
    //
    //        // Calculation
    //        double thisDistance;
    //        double x0;
    //        double x1;
    //        double y0;
    //        double y1;
    //        double value0;
    //        double value1;
    //        // diff: The sum of all the differences between the grid cells
    //        if ( comparator.equalsIgnoreCase( "diff" ) ) {
    //            double diff = 0.0d;
    //            for ( int i = 0; i < grid0Nrows; i ++ ) {
    //                for ( int j = 0; j < grid0Ncols; j ++ ) {
    //                    value0 = grid0.getCell( i, j );
    //                    if ( value0 != grid0NoDataValue ) {
    //                        value1 = grid0.getCell( i, j );
    //                        if ( value1 != grid1NoDataValue ) {
    //                            diff += ( value0 - value1 );
    //                        }
    //                    }
    //                }
    //            }
    //            return diff;
    //        }
    //
    //        // abs: The sum of all the absolute differences between the grid cells
    //        if ( comparator.equalsIgnoreCase( "abs" ) ) {
    //            double abs = 0.0d;
    //            for ( int i = 0; i < grid0Nrows; i ++ ) {
    //                for ( int j = 0; j < grid0Ncols; j ++ ) {
    //                    value0 = grid0.getCell( i, j );
    //                    if ( value0 != grid0NoDataValue ) {
    //                        value1 = grid0.getCell( i, j );
    //                        if ( value1 != grid1NoDataValue ) {
    //                            abs += Math.abs( value0 - value1 );
    //                        }
    //                    }
    //                }
    //            }
    //            return abs;
    //        }
    //
    //        // pearsons: The persons correlation coefficient
    //        if ( comparator.equalsIgnoreCase( "pearsons" ) ) {
    //            double pearsons = grid0NoDataValue;
    //            double sum0 = 0.0d;
    //            double sum1 = 0.0d;
    //            double sum0Squared = 0.0d;
    //            double sum1Squared = 0.0d;
    //            double sum01 = 0.0d;
    //            double n = 0.0d;
    //            for ( int i = 0; i < grid0Nrows; i ++ ) {
    //                for ( int j = 0; j < grid0Ncols; j ++ ) {
    //                    value0 = grid0.getCell( i, j );
    //                    value1 = grid1.getCell( i, j );
    //                    if ( value0 != grid0NoDataValue && value1 != grid0NoDataValue ) {
    //                        sum0 += value0;
    //                        sum1 += value1;
    //                        sum0Squared += Math.pow( value0, 2.0d );
    //                        sum1Squared += Math.pow( value1, 2.0d );
    //                        sum01 += value0 * value1;
    //                        n += 1.0d;
    //                    }
    //                }
    //            }
    //            double numerator = ( n * sum01 ) - ( sum0 * sum1 );
    //            double denominator = Math.sqrt( Math.abs( (  ( n * sum0Squared ) - Math.pow( sum0, 2.0d )  ) * ( ( n * sum1Squared ) - Math.pow( sum1, 2.0d ) ) ) );
    //            if ( denominator != 0.0d ) {
    //                pearsons = numerator / denominator;
    //            }
    //            return pearsons;
    //        }
    //
    //        // momentcorrelation: The moment correlation coefficient
    //        if ( comparator.equalsIgnoreCase( "momentCorrelation" ) ) {
    //            double momentCorrelation = grid0NoDataValue;
    //            double mean0 = grid0Statistics.getMean();
    //            double mean1 = grid1Statistics.getMean();
    //            double sd0 = 0.0d;
    //            double sd1 = 0.0d;
    //            double m01 = 0.0d;
    //            double n = 0.0d;
    //            double denominator;
    //            for ( int i = 0; i < grid0Nrows; i ++ ) {
    //                for ( int j = 0; j < grid0Ncols; j ++ ) {
    //                    value0 = grid0.getCell( i, j );
    //                    value1 = grid1.getCell( i, j );
    //                    if ( value0 != grid0NoDataValue && value1 != grid1NoDataValue ) {
    //                        n += 1.0d;
    //                        sd0 += Math.pow( value0 - mean0, 2.0d );
    //                        sd1 += Math.pow( value1 - mean1, 2.0d );
    //                        m01 += ( value0 - mean0 ) * ( value1 - mean1 );
    //                    }
    //                }
    //            }
    //            denominator = Math.sqrt( sd0 ) * Math.sqrt( sd1 );
    //            if ( denominator != 0.0d ) {
    //                momentCorrelation = m01 / denominator;
    //            }
    //            return momentCorrelation;
    //        }
    //        return grid0NoDataValue;
    //    }
    
    //    /**
    //     * Returns a new Grid2DSquareCellDouble the values of which are the distance to the nearest data value
    //     * TODO: Optimise as it is currently very slow and inefficient!!!
    //     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
    //     *   in this method then swap operations are initiated prior to retrying.
    //     *   If false then OutOfMemoryErrors are caught and thrown.
    //     */
    //    public Grid2DSquareCellDouble distanceToDataValue( Grid2DSquareCellDouble grid0, Grid2DSquareCellDoubleFactory gridFactory, boolean handleOutOfMemoryError ) {
    //        long nrows = grid0.getNrows();
    //        long ncols = grid0.getNcols();
    //        BigDecimal[] dimensions0 = grid0.getDimensions();
    //        double noDataValue = grid0.getNoDataValue();
    //        Grid2DSquareCellDouble result = gridFactory.create( grid0.getChunkNrows(this.handleOutOfMemoryErrorFalse), grid0.getChunkNcols(this.handleOutOfMemoryErrorFalse), nrows, ncols, dimensions0, noDataValue );
    //        // Calculate distances
    //        long row;
    //        long col;
    //        for ( row = 0; row < nrows; row ++ ) {
    //            for ( col = 0; col < ncols; col ++ ) {
    //                result.setCell( row, col, grid0.getNearestValueDoubleDistance( row, col, this.handleOutOfMemoryErrorTrue ), this.handleOutOfMemoryErrorTrue );
    //            }
    //        }
    //        return result;
    //    }
    
    //    TODO:
    //    /**
    //     * Returns a new Grid2DSquareCellDouble which is a copy of grid0 but
    //     * with all noDataValues replaced as follows:
    //     * At each iteration we deal with the nearest band of noDataValues.
    //     * (NB. If we dealt only with the nearest noDataValues it may take a very long time to compute!)
    //     * We replace these nearest noDataValues with the average values within distance.
    //     * We then repeat until all cells have a value.
    //     */
    //    public Grid2DSquareCellDouble replaceNoDataValues( Grid2DSquareCellDouble grid0, double distance, Grid2DSquareCellDoubleFactory gridFactory ) {
    //        long nrows = grid0.getNrows();
    //        long ncols = grid0.getNcols();
    //        BigDecimal[] dimensions = grid0.getDimensions();
    //        double noDataValue = grid0.getNoDataValue();
    //        Grid2DSquareCellDouble temp1 = gridFactory.createGrid2DSquareCellDouble( grid0 );
    //        Grid2DSquareCellDouble temp2 = gridFactory.createGrid2DSquareCellDouble( grid0 );
    //        // Get distances
    //        Grid2DSquareCellDouble distanceGrid = distanceToDataValue( grid0, gridFactory );
    //        AbstractGridStatistics distanceGridStatistics = distanceGrid.getGridStatistics();
    //        double maxDistance = distanceGridStatistics.getMax();
    //
    //        AbstractGrid2DSquareCellDouble[] geometricDensity = Grid2DSquareCellDoubleProcessorGWS.geometricDensity( grid0, maxDistance, gridFactory );
    //        //return geometricDensity[ geometricDensity.length - 1 ];
    //        return geometricDensity[ geometricDensity.length / 2 ];
    //        /*
    //        int maxCellDistance = ( int ) Math.ceil( maxDistance / cellsize );
    //        boolean alternator = true;
    //        Grid2DSquareCellDouble thisGrid = null;
    //        Grid2DSquareCellDouble thatGrid = null;
    //        double thisValue;
    //        double thatValue;
    //        double thisDistance;
    //        for ( int iterations = 1; iterations <= maxCellDistance; iterations ++ ) {
    //            for ( int i = 0; i < nrows; i ++ ) {
    //                for ( int j = 0; j < ncols; j ++ ) {
    //                    thisValue = thisGrid.getCell( i, j );
    //                    thatValue = thatGrid.getCell( i, j );
    //                    thisDistance = distanceGrid.getCell( i, j );
    //                    if ( thatValue != noDataValue ) {
    //                        thisGrid.setCell( i, j, thatValue );
    //                    } else {
    //                        if ( thatValue == noDataValue && thisDistance < ( iterations * cellsize ) ) {
    //                            thisGrid.setCell( i, j, Grid2DSquareCellDoubleProcessor.regionUnivariateStatistics( thatGrid, i, j, "mean", distance, 1.0d, 1.0d, gridFactory ) );
    //                        }
    //                    }
    //                }
    //            }
    //        }
    //        thatGrid.clear();
    //        return thisGrid;
    //         */
    //
    //    }
    
    
    /**
     * Shuffles dummyDiff numberOfShuffles places
     */
    private void shuffle( double[] dummyDiff, double[] diff, int numberOfShuffles ) {
        for ( int i = 0; i < dummyDiff.length - numberOfShuffles; i ++ ) {
            dummyDiff[ i ] = diff[ i + numberOfShuffles ];
        }
        for ( int i = dummyDiff.length - 1; i > dummyDiff.length - 1 + numberOfShuffles; i -- ) {
            dummyDiff[ i ] = diff[ i - dummyDiff.length + numberOfShuffles ];
        }
    }
    
    /**
     * Returns a double[][] of grid values
     *
     */
    protected double[][] getRowProcessInitialData( 
            Grid2DSquareCellDouble grid, 
            int cellDistance,
            long row ) {
        double[][] result = new double[ ( cellDistance * 2 ) + 1 ][ ( cellDistance * 2 ) + 1 ];
        long col;
        long r;
        for ( r = - cellDistance; r <= cellDistance; r ++ ) {
            for ( col = - cellDistance; col <= cellDistance; col ++ ) {
                result[ ( int ) row + cellDistance ][ ( int ) col + cellDistance ] = 
                        grid.getCell( r + row, col, this.handleOutOfMemoryErrorTrue );
            }
        }
        return result;
    }
    
    /**
     * Returns a double[][] based on previous which has been shuffled
     */
    protected double[][] getRowProcessData( 
            Grid2DSquareCellDouble grid, 
            double[][] previous, 
            int cellDistance, 
            long row, 
            long col ) {
        double[][] result = previous;
        if ( col == 0 ) {
            return getRowProcessInitialData( 
                    grid, 
                    cellDistance, 
                    row );
        } else {
            // shift columns one left
            for ( int i = 0; i <= cellDistance * 2; i ++ ) {
                for ( int j = 0; j <= ( cellDistance * 2 ) - 1; j ++ ) {
                    result[ i ][ j ] = previous[ i ][ j + 1 ];
                }
            }
            // getLastColumn
            for ( int i = -cellDistance; i <= cellDistance; i ++ ) {
                result[ i + cellDistance ][ cellDistance * 2 ] = 
                        grid.getCell( 
                        ( long ) i + row, 
                        ( long ) col + cellDistance, 
                        this.handleOutOfMemoryErrorTrue );
            }
        }
        return result;
    }
    
    
    //    // TODO:
    //    // tests
    //    private double[][] getChunkProcessInitialData( Grid2DSquareCellDouble grid, int chunkCells, int rowChunk ) {
    //        double[][] result = new double[ chunkCells * 3 ][ chunkCells * 3 ];
    //        for ( int i = -chunkCells; i <= ( chunkCells * 2 ) - 1; i ++ ) {
    //            for ( int j = -chunkCells; j <= ( chunkCells * 2 ) - 1; j ++ ) {
    //                result[ i + chunkCells ][ j + chunkCells ] = grid.getCell( i + ( rowChunk * chunkCells ), j );
    //            }
    //        }
    //        return result;
    //    }
    //
    //    // Needs testing!!
    //    private double[][] getChunkProcessData( Grid2DSquareCellDouble grid, double[][] previous, int chunkCells, int rowChunk, int colChunk ) {
    //        double[][] result = previous;
    //        if ( colChunk == 0 ) {
    //            return getRowProcessInitialData( grid, chunkCells, rowChunk );
    //        } else {
    //            // shift end columns to start columns
    //            for ( int i = -chunkCells; i <= ( chunkCells * 2 ) - 1; i ++ ) {
    //                for ( int j = 0; j <= chunkCells; j ++ ) {
    //                    result[ i + chunkCells ][ j ] = previous[ i + ( chunkCells * 2 ) ][ j + ( chunkCells * 2 ) ];
    //                }
    //            }
    //            // getOtherData
    //            for ( int i = -chunkCells; i <= ( chunkCells * 2 ) - 1; i ++ ) {
    //                for ( int j = 0; i <= ( chunkCells * 2 ) - 1; i ++ ) {
    //                    result[ i + chunkCells ][ j + chunkCells ] = grid.getCell( i + ( rowChunk * chunkCells ), j + ( colChunk * chunkCells ) );
    //                }
    //            }
    //        }
    //        return result;
    //    }
    
}